import json, re, threading

from datetime import datetime, timedelta
from collections import OrderedDict
from functools import update_wrapper
from hashlib import sha1

from .dotdict import DotDict

from typing import *


__all__ = [
	'BaseCache',
	'TtlCache',
	'LruCache'
]


def parse_ttl(ttl):
	if not ttl:
		return 0

	if type(ttl) == int:
		return ttl

	if type(ttl) == timedelta:
		return timedelta.seconds

	raise TypeError(f'ttl must be an int or timedelta object, not {type(ttl).__name__}')


## This doesn't work for some reason
def CacheDecorator(cache):
	def decorator(func):
		def wrapper(cls, *args, **kwargs):
			key = sha1(json.dumps(args).encode() + json.dumps(kwargs).encode()).hexdigest()
			cached = cache.fetch(key)
			print(cached)

			if cached != None:
				print('Returning cached value:', cache)
				return cached

			result = func(cls, *args, **kwargs)

			cache.store(key, result)

			return result

		return wrapper
	return decorator


class DefaultValue(object):
	pass


class BaseCache(OrderedDict):
	'Base cache class'

	def __init__(self, maxsize:int=1024, ttl:Union[int,timedelta]=None):
		self.ttl = parse_ttl(ttl)
		self.maxsize = maxsize
		self.fetch = self.get
		self.store = self.set

		self.deco = lambda *args: CacheDecorator(self, *args)

		# This lock hasn't been tested yet, but it should work
		self._lock = threading.Lock()


	def __str__(self):
		return f'BaseCache(maxsize={self.maxsize}, ttl={self.ttl})'


	def _check_size(self):
		if not self.maxsize:
			return

		while True:
			if len(self) < self.maxsize:
				break

			self.popitem(last=False)


	def clear(self):
		'Remove all data from the cache'
		for key in list(self.keys()):
			self.remove(key)


	def items(self) -> dict:
		'Return cached items as a dict'
		return [[k, v.data] for k,v in super().items()]


	def remove(self, key:str):
		'Remove an item by key'

		with self._lock:
			if self.get(key):
				del self[key]


	def set(self, key:str, value:Any) -> Any:
		'Store a new key/value pair'

		with self._lock:
			if not self.get(key):
				self[key] = DotDict()

			self[key].data = value

			if self.ttl:
				timestamp = int(datetime.timestamp(datetime.now()))
				self[key].timestamp = timestamp + self.ttl

			self.move_to_end(key)


	def get(self, key:str):
		'Fetch an item by key'

		item = super().get(key)

		if not item:
			return

		if self.ttl:
			timestamp = int(datetime.timestamp(datetime.now()))

			if timestamp >= self[key].timestamp:
				del self[key]
				return

			self[key]['timestamp'] = timestamp + self.ttl

		self.move_to_end(key)

		return item['data']


	def fetch(self, key:str) -> Any:
		'Alias for BaseCache.get'

		return self.get(key)


	def store(self, key:str, value:Any):
		'Alias for BaseCache.set'

		return self.set(key, value)


class TtlCache(BaseCache):
	'Cache that auto-removes pairs after a set amount of time'

	def __init__(self, ttl:Union[int,timedelta]=timedelta(hours=1)):
		super().__init__(maxsize, ttl)


class LruCache(BaseCache):
	'Cache that auto-removes pairs after a set item limit'

	def __init__(self, maxsize:int=1024):
		super().__init__(maxsize)
