from colorsys import rgb_to_hls, hls_to_rgb
from functools import cached_property, partial
from izzylib import DotDict


pride_colors = {
	'lgbt': ['#9400D3', '#4B0082', '#0000FF', '#00FF00', '#FFFF00', '#FF7F00', '#FF0000'],
	'lesbian': ['#D52D00', '#EF7627', '#FF9A56', '#FFFFFF', '#D162A4', '#B55690', '#A30262'],
	'bi': ['#D60270', '#9B4F96', '#0038A8'],
	'gay': ['#078D70', '#26CEAA', '#98E8C1', '#FFFFFF', '#7BADE2', '#5049CC', '#3D1A78'],
	'pan': ['#FF218C', '#FFD800', '#21B1FF'],
	'asexual': ['#000000', '#A3A3A3', '#FFFFFF', '#800080'],
	'aromantic': ['#3DA542', '#A7D379', '#FFFFFF', '#A9A9A9', '#000000'],
	'trans': ['#55CDFC', '#F7A8B8', '#FFFFFF'],
	'trans-black': ['#55CDFC', '#F7A8B8', '#000000'],
	'transmasc': ['#FF8ABD', '#CDF5FE', '#9AEBFF', '#74DFFF'],
	'non-binary': ['#FCF434', '#FFFFFF', '#9C59D1', '#2C2C2C']
}


class Color(str):
	def __new__(cls, color: str):
		if not isinstance(color, str):
			raise TypeError('Color must be a valid hex string')

		if not color.startswith('#'):
			color = f'#{str(color)}'

		if len(color) == 4:
			color = f'#{color[1]*2}{color[2]*2}{color[3]*2}'

		elif len(color) != 7:
			raise TypeError('Color must be 3 or 6 character hex string')

		return str.__new__(cls, color)


	def __repr__(self):
		return f'Color({self})'


	@classmethod
	def new_from_rgb(cls, red, green, blue):
		values = [red, green, blue]

		for value in values:
			if 0 < value > 255:
				raise ValueError('Color value must be anywhere from 0 to 255')

		return cls('#' + ''.join(hex(value)[2:].rjust(2, '0') for value in values))


	@classmethod
	def new_from_hsl(cls, hue, saturation, luminance):
		hsl_values = [hue, saturation, luminance]

		for idx, value in enumerate(hsl_values):
			if 0 < value > 255:
				raise ValueError('HSV values must be anywhere from 0 to 255')

			hsl_values[idx] = value / 255

		values = [int(value * 255) for value in hls_to_rgb(hsl_values[0], hsl_values[2], hsl_values[1])]

		return cls('#' + ''.join(hex(value)[2:].rjust(2, '0') for value in values))


	@classmethod
	def new_pride_palette(cls, flag: str):
		'Returns a `tuple` of `Color` objects which represents a pride flag color palette'

		if flag == 'ace':
			flag = 'asexual'

		elif flag == 'aro':
			flag = 'aromantic'

		elif flag in ['enby', 'nonbinary']:
			flag = 'non-binary'

		return tuple(cls(color) for color in pride_colors[flag])


	@cached_property
	def rgb(self):
		return (self.red, self.green, self.blue)


	@cached_property
	def hsl(self):
		rgb = [value / 255 for value in self.rgb]
		values = [int(value * 255) for value in rgb_to_hls(*rgb)]
		return (values[0], values[2], values[1])


	@cached_property
	def red(self):
		return int(self[1:3], 16)


	@cached_property
	def green(self):
		return int(self[3:5], 16)


	@cached_property
	def blue(self):
		return int(self[5:7], 16)


	@cached_property
	def hue(self):
		return self.hsl[0]


	@cached_property
	def saturation(self):
		return self.hsl[1]


	@cached_property
	def lightness(self):
		return self.hsl[2]


	def alter(self, action, multiplier):
		hue, saturation, luminance = self.hsl

		if action == 'lighten':
			luminance += (255 - luminance) * parse_multi(multiplier)

		elif action == 'darken':
			luminance -= luminance * parse_multi(multiplier)

		elif action == 'saturate':
			saturation += (255 - saturation) * parse_multi(multiplier)

		elif action == 'desaturate':
			saturation -= saturation * parse_multi(multiplier)

		else:
			raise ValueError(f'Invalid action: {action}')

		return self.__class__.new_from_hsl(hue, saturation, luminance)


	def lighten(self, multiplier):
		return self.alter('lighten', multiplier)


	def darken(self, multiplier):
		return self.alter('darken', multiplier)


	def saturate(self, multiplier):
		return self.alter('saturate', multiplier)


	def desaturate(self, multiplier):
		return self.alter('desaturate', multiplier)


	def rgba(self, opacity):
		if 0 < opacity > 100:
			raise ValueError('Opacity must anywhere from 0 to 100')

		values = ', '.join(str(value) for value in self.rgb)
		trans = opacity / 100
		return f'rgba({values}, {trans:.2})'


def parse_multi(multiplier):
	if multiplier >= 100:
		return 1

	elif multiplier <= 0:
		return 0

	return multiplier / 100


lighten = lambda color, multi: Color(color).alter('lighten', multi)
darken = lambda color, multi: Color(color).alter('darken', multi)
saturate = lambda color, multi: Color(color).alter('saturate', multi)
desaturate = lambda color, multi: Color(color).alter('desaturate', multi)
rgba = lambda color, multi: Color(color).rgba(multi)
