from collections.abc import Callable
from typing import *

from . import izzylog
from .dotdict import DotDict
from .path import Path

try:
	import yaml
except ImportError:
	izzylog.verbose('Failed to import pyyaml. YamlConfig disabled')
	yaml = None


__all__ = [
	'BaseConfig',
	'JsonConfig',
	'YamlConfig'
]


class BaseConfig(DotDict):
	'''
		DotDict that only allows certain keys to be set and can load/save to a file. This is determined by the defaults.

		To implement your own config format, create a sub-class and re-implement "_load_config" and "_save_config".

		Arguments:

		- config_file: Path to the config file.

		- value_parser: Function used to parse key/value pairs about to be added.

		- defaults: Key/value pairs to set on init and will determine what keys can be set later.
	'''

	def __init__(self,
			config_file:Union[Path,str]=None,
			value_parser:Callable[str,Any]=None,
			**defaults:Dict[str,Any]
		):
		self._defaults = defaults
		self._config_file = Path(config_file)
		self._parser = value_parser or self._parse_value

		super().__init__(defaults)


	def __setitem__(self, key, value):
		if not key in self._defaults:
			raise KeyError(f'Not a valid config option: {key}')

		value = self._parser(key, value)
		super().__setitem__(key, value)


	def _load_config(self):
		'Reads the config and returns a "dict".'
		pass


	def _save_config(self):
		'Takes the current config and stores it in the file specified by "config_file".'
		pass


	def _parse_value(self, key, value):
		'Does any necessary processing of a value before it gets set.'
		return value


	def update(self, _data={}, **kwargs):
		if _data == None:
			return

		data = {**_data, **kwargs}

		for k, v in data.items():
			self[k] = v


	def reset(self, key=None):
		'Sets every key to the default value.'
		if not key:
			self.update(self._defaults)

		else:
			self[key] = self._defaults[key]


	def load(self):
		'Imports a config file.'
		self.update(self._load_config() or {})


	def save(self):
		'Saves the current config to a file and makes the parent directory if necessary.'
		self._config_file.parent.mkdir()
		self._save_config()


class JsonConfig(BaseConfig):
	'''Config class that uses JSON as the backing config.'''


	def _load_config(self):
		self.load_json(self._config_file)


	def _save_config(self):
		self.save_json(self._config_file, indent='\t')


class YamlConfig(BaseConfig):
	'''Config class that uses YAML as the backing config. If PyYaml is not installed, an ImportError will be raised on init.'''


	def __init__(self, *args, **kwargs):
		if not yaml:
			raise ImportError('PyYaml not found')

		super().__init__(*args, **kwargs)


	def _load_config(self):
		with self._config_file.open('r') as fd:
			data = yaml.safe_load(fd)

		return data


	def _save_config(self):
		data = dict(self)

		for k,v in data.items():
			if isinstance(v, dict):
				data[k] = dict(v)

		with self._config_file.open('w') as fd:
			yaml.safe_dump(data, fd, sort_keys=False)
