import enum
import json
import socket
import threading

from collections.abc import Callable
from typing import *

from . import izzylog
from .dotdict import DotDict
from .misc import convert_to_bytes


__pdoc__ = {
	'BaseConnection.address': 'The address to be used for this connection',
	'BaseConnection.decode': 'Whether or not to convert incoming messages to a str object',
	'BaseConnection.encoding': 'The encoding to use when decoding bytes',
	'BaseConnection.port': 'The port to be used for this connection',
	'Protocol': 'Values to specify the type of connection being made.',
	'Protocol.BROADCAST': 'Specifies a UDP broadcast connection',
	'Protocol.MULTICAST': 'Specifies a UDP multicast connection',
	'Protocol.TCP': 'Specifies a TCP connection',
	'Prococol.UDP': 'Specifies a UDP connection'
}


__all__ = [
	'BaseConnection',
	'Client',
	'ClientConnection',
	'Protocol',
	'Server'
]


Protocol = enum.Enum('Protocol', [
		'TCP',
		'UDP',
		'BROADCAST',
		'MULTICAST'
	]
)


class BaseConnection(socket.socket):
	'''
		[WIP] An easy to use wrapper around the socket module for TCP and UDP connections. Can be used like a context manager for client connections.

		Arguments:

		- address: v4 or v6 IP address

		- port: TCP or UDP port to use

		- proto: The protocol to use.

		- timeout: How long to wait when recieving data before raising a TimeoutError
	'''

	#__slots__ = ['address', 'port', 'encoding', 'decode', '_proto', '_timeout', '_ttl', '_connected']

	address = '127.0.0.1'
	port = 8080

	_protocol = Protocol.TCP
	_timeout = 80
	_ttl = 1
	_connected = False


	def __init__(self, address=None, port=None, **kwargs):
		if address:
			self.address = address

		if port:
			self.port = int(port)

		for key, value in kwargs.items():
			self.set_property(key, value)

		if self.protocol == Protocol.TCP:
			super().__init__(socket.AF_INET, socket.SOCK_STREAM)

		else:
			super().__init__(socket.AF_INET, socket.SOCK_DGRAM)


	def __enter__(self):
		self.connect()
		return self


	def __exit__(self, exctype, value, tb):
		self.disconnect()


	@property
	def connected(self) -> bool:
		'Whether or not a client is connected or a server is listening'
		return self._connected


	@property
	def protocol(self) -> Protocol:
		'The protocol used for this connection'
		return self._protocol


	@protocol.setter
	def protocol(self, value):
		if type(value) == str:
			self._proto = Protocol[value.upper()]

		elif type(value) != Protocol:
			raise TypeError(f'proto must be a Protocol enum or a string')

		else:
			self._proto = value


	@property
	def timeout(self) -> int:
		'How long to wait when recieving data before raising a TimeoutError'
		return self._timeout


	@timeout.setter
	def timeout(self, value:int):
		'Set the timeout value'
		self._timeout = value
		self.settimeout(value)


	@property
	def ttl(self) -> int:
		'heck'
		return self._ttl


	@ttl.setter
	def ttl(self, value: int):
		self._ttl = value

		if self.config.proto == protocol.BROADCAST:
			self.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, ttl)

		elif self.config.proto == protocol.MULTICAST:
			self.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, ttl)


	def set_property(self, key, value):
		self.__setattr__(key, value)


	def _connect(self):
		pass


	def _disconnect(self):
		pass


	def connect(self):
		'Connect to a remote socket as a client. Will silently return if already connected.'

		if self._connected:
			return self

		izzylog.verbose('Opening connection')

		try:
			self._connected = True
			self._connect()

		except Exception as e:
			self._connected = False
			raise e from None

		return self


	def disconnect(self):
		'Close the connection to a remote socket. Will silently return if not connected.'

		izzylog.verbose('Closing connection')

		self.shutdown(socket.SHUT_RDWR)
		self.close()


class Client(BaseConnection):
	'Creates a connection to a remote server'


	def __str__(self):
		return f'<Client {self.address}:{self.port} ({self.protocol.name})>'


	def _connect(self):
		self.setblocking(False)
		socket.socket.connect(self, (self.address, self.port))


	def read(self, size:int=8192, decode=False, encoding='utf-8'):
		'Read incoming data from the server.'

		if not self.connected:
			raise ConnectionError('Not connected')

		data = self.recv(size)
		return data.decode(encoding) if decode else data


	def write(self, msg:Any):
		'Send data to a server. If the message is not a bytes object, it will attempt to be converted to one.'

		if not self.connected:
			raise ConnectionError('Not connected')

		self.sendall(convert_to_bytes(msg))


	def write_lines(self, *lines):
		data = b''

		for line in lines:
			data += convert_to_bytes(line) + b'\r\n'

		self.write(data)


class Server(BaseConnection):
	'Accepts connections from remote clients. Can iterate over the server to return client connections.'


	def __init__(self, *args, backlog:int=100, client_handler:Callable=None, client_class=None, **kwargs):
		super().__init__(*args, **kwargs)

		self._backlog = backlog

		self._client_class = client_class or ClientConnection
		self._client_handler = client_handler
		self.connections = {}
		self.clients = ClientThreads(self)

		self.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)


	def __str__(self):
		return f'<Server {self.address}:{self.port} ({self.protocol.name})>'


	def __iter__(self):
		self.connect()

		while True:
			if not self.connected:
				break

			yield self._client_class(self, *self.accept())


	def _connect(self):
		self.bind((self.address, self.port))
		self.listen(self._backlog)


	@property
	def backlog(self):
		'The number of connections to place in the queue before rejecting new ones'

		return self._backlog


	def run(self):
		'Start listening on the specified address and port and handle incoming clients'

		self.connect()

		for client in self:
			self.clients.add(client)


class ClientConnection(threading.Thread):
	'Object that represents a client connection on a server'


	def __init__(self, server, connection, address):
		super().__init__()

		self.server = server
		self.connection = connection
		self.address = address

		connection.setblocking(False)


	def __enter__(self):
		return self


	def __exit__(self, *args):
		self.close()


	def close(self):
		self.close()
		self.server.clients.remove(self)


	def read(self, size:int=8192):
		'Read incoming data from the server.'
		try:
			return self.connection.recv(size)
		except BlockingIOError:
			return b''


	def write(self, msg:Any):
		if isinstance(msg, str):
			msg = msg.encode('utf-8')

		self.connection.sendall(msg)


	def write_lines(self, *lines):
		data = b''

		for line in lines:
			data += convert_to_bytes(line) + b'\r\n'

		self.write(data)


	def run(self):
		self._server._client_handler(self)


class ClientThreads(list):
	def __init__(self, server):
		super().__init__()

		self._server = server
		self._lock = threading.Lock()


	def add(self, client):
		with self._lock:
			client.start()
			super().append(client)


	def remove(self, client):
		with self._lock:
			client.close()
			super().remove(client)



