from datetime import datetime, timezone

from .enums import DatetimeFormat


__all__ = ['DateString']
__pdoc__ = {
	'DateString.dt': '`datetime.datetime` object used to store the date',
	'DateString.format': 'The format used to create the str',
	'DateString.tz_local': 'The local timezone as a `datetime.timezone` object',
	'DateString.tz_utc': 'UTC timezone as a `datetime.timezone` object',
}


class DateString(str):
	'Create a `str` based on a datetime object'

	tz_utc = timezone.utc
	tz_local = datetime.now(tz_utc).astimezone().tzinfo
	dt = None
	format = None


	def __init__(self, string, format):
		self.format = format if isinstance(format, DatetimeFormat) else DatetimeFormat[format.upper()]

		try:
			self.dt = datetime.strptime(string, self.format.value).replace(tzinfo=self.tz_utc)

		except ValueError:
			if format.upper() == 'ACTIVITYPUB':
				self.dt = datetime.strptime(string, DatetimeFormat['ACTIVITYPUB_MICRO'].value).replace(tzinfo=self.tz_utc)


	def __new__(cls, string, format):
		date = str.__new__(cls, string)
		return date


	def __getattr__(self, key):
		return getattr(self.dt, key)


	def __repr__(self):
		return f'DateString({self}, format={self.format})'


	@classmethod
	def new_activitypub(cls, date):
		'Create a new `DateString` for use in ActivityPub or Mastodon API messages'
		return cls(date, 'activitypub')


	@classmethod
	def new_http(cls, date):
		'Create a new `DateString` for use in HTTP messages'
		return cls(date, 'http')


	@classmethod
	def new_mastodon(cls, date):
		return cls(date, 'mastodon')


	@classmethod
	def from_datetime(cls, date, format):
		'Create a new `DateString` from an existing `datetime.datetime` object'
		str_format = format if isinstance(format, DatetimeFormat) else DatetimeFormat[format.upper()]

		if not date.tzname():
			date = date.astimezone(cls.tz_utc)

		return cls(date.strftime(str_format.value), format)


	@classmethod
	def from_timestamp(cls, timestamp, format):
		'Create a new `DateString` from a unix timestamp'
		return cls.from_datetime(datetime.fromtimestamp(timestamp), format)


	@classmethod
	def now(cls, format):
		'Create a new `DateString` from the current time'
		return cls.from_datetime(datetime.now(cls.tz_utc), format)


	@property
	def activitypub(self):
		'Return a new `DateString` in ActivityPub format'
		return DateString(self.dump_to_string('activitypub'), 'activitypub')


	@property
	def http(self):
		'Return a new `DateString` in HTTP format'
		return DateString(self.dump_to_string('http'), 'http')


	@property
	def mastodon(self):
		'Return a new `DateString` in Mastodon API format'
		return DateString(self.dump_to_string('mastodon'), 'mastodon')


	@property
	def utc(self):
		'Return a new `DateString` in UTC time'
		return DateString.from_datetime(self.dt.astimezone(self.tz_utc), self.format)


	@property
	def local(self):
		'Return a new `DateString` in local time'
		return DateString.from_datetime(self.dt.astimezone(self.tz_local), self.format)


	def dump_to_string(self, format=None):
		'Return the date as a normal string in the specified format'
		if not format:
			format = self.format

		str_format = format if isinstance(format, DatetimeFormat) else DatetimeFormat[format.upper()]
		return self.dt.strftime(format.value)
