import copy, json

from urllib.parse import unquote_plus, quote

from .path import Path

from typing import *


__all__ = [
	'CapitalDotDict',
	'DotDict',
	'DefaultDotDict',
	'LowerDotDict',
	'MultiDotDict',
	'JsonEncoder'
]

__pdoc__ = {
	'JsonEncoder.__init__': False
}


class DotDict(dict):
	'Python dictionary, but keys can be accessed, set, or deleted via attributes. Can also directly parse JSON on init.'

	_ignore_keys = []
	_encoder = None

	def __init__(self, _value:dict={}, ignore_keys:list=[], encoder=None, **kwargs):
		super().__init__()

		self._ignore_keys = ignore_keys
		self._encoder = encoder or JsonEncoder

		if isinstance(_value, (str, bytes)):
			self.from_json(_value)

		elif isinstance(_value, dict):
			self.update(_value)

		elif isinstance(_value, (list, set)):
			self.from_pairs(_value)

		elif _value:
			raise TypeError(f'The value must be a JSON string, list, set, dict, or another DotDict object, not {_value.__class__}')

		if kwargs:
			self.update(kwargs)


	def __getattr__(self, k):
		try:
			return super().__getitem__(k)

		except KeyError:
			raise AttributeError(f'{self.__class__.__name__} object has no attribute {k}') from None


	def __setattr__(self, k, v):
		try:
			if k in self._ignore_keys:
				super().__setattr__(k, v)

		except AttributeError:
			pass

		if k.startswith('_'):
			super().__setattr__(k, v)

		else:
			self.__setitem__(k, v)


	def __setitem__(self, k, v):
		if type(v) == dict:
			v = DotDict(v)

		super().__setitem__(k, v)


	def __delattr__(self, k):
		try:
			dict.__delitem__(self, k)
		except KeyError:
			raise AttributeError(f'{self.__class__.__name__} object has no attribute {k}') from None


	@classmethod
	def new_from_json_file(cls, path):
		'Create a new `DotDict` by loading a JSON file specified by `path`'

		data = cls()
		data.load_json(path)
		return data


	@classmethod
	def new_from_query_string(cls, query):
		'Create a new `DotDict` by parsing an HTTP query string'

		data = cls()
		data.from_query(query)
		return data


	@classmethod
	def new_from_fd(cls, fd):
		'Create a new `DotDict` by reading a file descriptor or `Path` object that points to a JSON file'

		with fd:
			return cls(fd.read())


	@property
	def ignore_keys(self):
		'Keys to be set as attributes instead of items'
		return self._ignore_keys


	def copy(self):
		'Creates a deep copy of the dict'

		return DotDict(copy.deepcopy(self))


	def update(self, data:dict={}, **kwargs):
		'Adds (or updates) new key/value pairs to the dict'

		kwargs.update(data)

		for k,v in kwargs.items():
			self.__setitem__(k, v)


	def from_pairs(self, data:List[List[Any]]):
		'Updates dict from a list or set of list or set pairs'

		for key, value in data:
			self[key] = value


	def from_query(self, query:str):
		'Updates dict from HTTP query string'

		if not query:
			return

		for qline in query.split('&'):
			try: key, value = qline.split('=', 1)
			except: key, value = qline, None

			self[key] = unquote_plus(value)


	def to_query(self) -> str:
		'Returns dict as an HTTP query string'

		return quote('&'.join(f'{quote(k)}={quote(v)}' for k,v in self.items()))


	def to_json(self, indent:Union[int,str]=None, **kwargs) -> str:
		'Returns the dict as a JSON string. `indent` and `kwargs` are passed to `json.dumps`'

		encoder = kwargs.pop('cls', self._encoder)
		return json.dumps(self, indent=indent, cls=encoder, **kwargs)


	def from_json(self, string):
		'Updates the dict from a JSON string'

		data = json.loads(string)
		self.update(data)


	def load_json(self, path):
		'Updates the dict from a JSON file'

		with open(path) as fd:
			self.update(json.load(fd))


	def save_json(self, path, indent=None):
		'Dumps the dict to a file as JSON'

		with open(path, 'w') as fp:
			fp.write(self.to_json(indent=indent))


class DefaultDotDict(DotDict):
	'A `DotDict`, but accessing non-existent keys creates a new `DefaultDotDict` as a value for that key'

	def __getitem__(self, key):
		try:
			value = super().__getitem__(key)

		except KeyError:
			value = DefaultDotDict()
			super().__setitem__(self, key, value)

		return value


class LowerDotDict(DotDict):
	'A `DotDict`, but all keys are converted to lowercase'

	def __getitem__(self, key):
		return super().__getitem__(key.lower())


	def __setitem__(self, key, value):
		return super().__setitem__(key.lower(), value)


	def __delitem__(self, key):
		return super().__delitem__(key.lower())


class CapitalDotDict(DotDict):
	'A `DotDict`, but all keys are capitalized'

	def __getitem__(self, key):
		return super().__getitem__(key.capitalize())


	def __setitem__(self, key, value):
		return super().__setitem__(key.capitalize(), value)


	def __delitem__(self, key):
		return super().__delitem__(key.capitalize())


class MultiDotDict(LowerDotDict):
	'A `DotDict` with support for multiple values per key'

	def __setitem__(self, key, value, single=False):
		key = key.lower()

		if single:
			super().__setitem__(key, [value])

		else:
			try:
				self.__getitem__(key, False).append(value)

			except KeyError as e:
				super().__setitem__(key, [value])


	def __getitem__(self, key, single=True):
		key = key.lower()
		values = super().__getitem__(key)

		if single:
			try:
				return values[0]

			except IndexError:
				return None

		return values


	def getall(self, key:Any, default:list=[]) -> List[Any]:
		'Get all values for a key. Return the default value if set'

		try:
			return super().__getitem__(key)

		except KeyError as e:
			if not isinstance(default, list):
				default = [default]

			return default


	def set(self, key, value):
		'Drop all values of a key and set it to the specified value'

		self.__setitem__(key, value, True)


	def delone(self, key, value):
		'Delete one value of a key'

		self.__getitem__(key, False).remove(value)


	def delete(self, key):
		'Delete all values for a key'

		del self[key]


class ReadOnlyDotDict(DotDict):
	__readonly = False


	def __init__(self, *args, **kwargs):
		DotDict.__init__(self, *args, **kwargs)

		self.__readonly = True


	def __setitem__(self, key, value):
		if self.__readonly:
			raise RuntimeError('Cannot set items on read-only dict')

		DotDict.__setitem__(self, key, value)


	def __delitem__(self, key):
		if self.__readonly:
			raise RuntimeError('Cannot set items on read-only dict')

		DotDict.__delitem__(self, key)


class JsonEncoder(json.JSONEncoder):
	'Class used to convert objects into JSON-compatible objects'

	# Only implemented to disable the docstring. There's probably a better way to do this tbh
	def __init__(self, *args, **kwargs):
		''
		super().__init__(*args, **kwargs)


	def default(self, obj):
		'The default encoding action'

		if not isinstance(obj, (str, int, float, dict)):
			return str(obj)

		return json.JSONEncoder.default(self, obj)
