from functools import partial


__all__ = []


def register(cls):
	__all__.append(cls.__name__)
	return cls


@register
class DNSResolverError(Exception):
	'Raise when a hostname failed to resolve'


@register
class InvalidAddressError(Exception):
	'Raise when a string is not actually an IPv4 Address'


@register
class NoConnectionError(Exception):
	'Raise when trying to use a connection that has not been established yet'


@register
class ReadOnlyError(Exception):
	'Raise when a read-only property is attempted to be set'


@register
class RouteFoundError(Exception):
	'Raise when attempting to add a router path that already exists'

	def __init__(self, path, method):
		Exception.__init__(self, f'{method.value} {path}')
		self.method = method
		self.path = path


@register
class RouteNotFoundError(Exception):
	'Raise when a Router path cannot be parsed'

	def __init__(self, path, method):
		Exception.__init__(self, f'{method.value} {path}')
		self.method = method
		self.path = path


@register
class TokenExistsError(Exception):
	'Raise when trying to create a token or app when a token already exists'


@register
class VersionError(Exception):
	'Raise when something is not the right version'


@register
class HttpError(Exception):
	def __init__(self, message, status=500):
		super().__init__(f'{message} {status}')

		self.status = status
		self.message = message


@register
class HttpClientError(HttpError):
	def __init__(self, response=None):
		super().__init__(response.status, response.reason)
		self.response = response


@register
class HttpRedirError(Exception):
	def __init__(self, path, status=301):
		super().__init__(f'HTTP Error {status}: {path}')

		self.status = status
		self.path = path


## 200 Errors
HttpOk = partial(HttpError, status=200)
HttpCreated = partial(HttpError,status=201)
HttpAccepted = partial(HttpError,status=202)
HttpNoContent = partial(HttpError,status=204)
HttpResetContent = partial(HttpError,status=205)
HttpPartialContent = partial(HttpError,status=206)

## 300 Errors
HttpNotModified = partial(HttpError, status=304)

## 400 Errors
HttpBadRequest = partial(HttpError, status=400)
HttpUnauthorized = partial(HttpError, status=401)
HttpForbidden = partial(HttpError, status=403)
HttpNotFound = partial(HttpError, status=404)
HttpMethodNotAllowed = partial(HttpError, status=405)
HttpRequestTimeout = partial(HttpError,status=408)
HttpGone = partial(HttpError,status=410)
HttpLengthRequired = partial(HttpError,status=411)
HttpPreconditionFailed = partial(HttpError,status=412)
HttpPayloadTooLarge = partial(HttpError,status=413)
HttpUriTooLong = partial(HttpError,status=414)
HttpUnsupportedMediaType = partial(HttpError,status=415)
HttpRangeNotSatisfiable = partial(HttpError,status=416)
HttpTeapot = partial(HttpError, status=418)
HttpUpgradeRequired = partial(HttpError,status=426)
HttpTooManyRequests = partial(HttpError,status=429)
HttpRequestHeaderFieldsTooLarge = partial(HttpError,status=431)
HttpUnavailableForLegalReasons = partial(HttpError,status=451)

## 500 Errors
HttpInternalServerError = partial(HttpError, status=500)
HttpNotImplemented = partial(HttpError,status=501)
HttpBadGateway = partial(HttpError,status=502)
HttpServiceUnavailable = partial(HttpError,status=503)
HttpGatewayTimeout = partial(HttpError,status=504)
HttpHttpVersionNotSuported = partial(HttpError,status=505)
HttpNetworkAuthenticationRequired = partial(HttpError,status=511)

## Redirects
HttpMovedPermanently = partial(HttpRedirError, status=301)
HttpFound = partial(HttpRedirError, status=302)
HttpSeeOther = partial(HttpRedirError, status=303)
HttpTemporaryRedirect = partial(HttpRedirError, status=307)
HttpPermanentRedirect = partial(HttpRedirError, status=309)
