from http.client import HTTPConnection, HTTPSConnection
from json import JSONDecodeError

from .request import HttpClientRequest
from .response import HttpClientResponse

from .. import activitypub
from .. import __version__, izzylog
from ..dotdict import DotDict
from ..exceptions import HttpClientError
from ..object_base import ObjectBase
from ..url import Url


class HttpClient(ObjectBase):
	def __init__(self, appagent=None, headers={}, request_class=None, response_class=None, timeout=60, redir_limit=10):
		super().__init__(
			headers = ClientHeaders(headers),
			timeout = timeout,
			redir_limit = redir_limit,
			request_class = request_class or HttpClientRequest,
			response_class = response_class or HttpClientResponse,
			readonly = ['headers']
		)

		self.set_appagent(appagent)


	@property
	def useragent(self):
		return self.headers.get('User-Agent')


	@useragent.setter
	def useragent(self, value):
		self.headers['User-Agent'] = value


	def build_request(self, url, method='GET', body=b'', headers={}, cookies=[], follow_redir=False):
		request = self.request_class(url,
			method = method,
			body = body,
			headers = self.headers.copy(),
			cookies = cookies,
			follow_redir = follow_redir
		)

		request.headers.update(headers)
		return request


	def request(self, url, method='GET', body=b'', headers={}, cookies=[], follow_redir=False, raise_error=False):
		req = self.build_request(url, method, body, headers, cookies, follow_redir)
		return self.send_request(req, raise_error)


	def send_request(self, request, raise_error=False):
		if request.url.proto == 'https':
			conn = HTTPSConnection(request.url.host, timeout=self.timeout)

		elif request.url.proto == 'http':
			conn = HTTPConnection(request.url.host, timeout=self.timeout)

		else:
			raise TypeError(f'Protocol not supported: {request.url.proto}')

		conn.request(request.method, request.url, request.body, request.headers)
		response = self.response_class(conn)
		origin = request.url
		redirs = []

		while response.status in range(300, 400) and request.follow_redir and response.location:
			response.close()
			assert len(redirs) < self.redir_limit, 'Reached redirect limit'

			redirs.append(response.location)

			request.url = response.location
			response = self.send_request(request)

		if raise_error and response.status not in range(200, 300):
			raise HttpClientError(response)

		response._origin = origin
		response._redirs = redirs
		return response


	def set_appagent(self, appagent=None):
		useragent = f'IzzyLib/{__version__}'

		if appagent:
			useragent += f' ({appagent})'

		self.set_useragent(useragent)


	def set_useragent(self, useragent=None):
		self.headers['User-Agent'] = useragent or f'IzzyLib/{__version__}'


	### Extra functions for specific purposes
	def fetch_wellknown_nodeinfo(self, domain):
		headers = {'Accept': 'application/json'}

		if isinstance(domain, Url):
			domain = url.hostname()

		with self.request(f'https://{domain}/.well-known/nodeinfo', headers=headers, raise_error=True) as resp:
			data = resp.text

			if not data:
				raise ValueError('Failed to fetch /.well-known/nodeinfo')

		return activitypub.WellknownNodeinfo(data)


	def fetch_nodeinfo(self, domain, version=None):
		'Fetch a domain\'s nodeinfo'

		headers = {'Accept': 'application/jrd+json'}

		## Fetch /.well-known/nodeinfo to find actual nodeinfo url
		wellknown = self.fetch_wellknown_nodeinfo(domain)
		nodeinfo_url = None

		if version:
			nodeinfo_url = wellknown.url(version)

		else:
			## Make sure a valid nodeinfo url exists
			for ni_version in activitypub.Nodeinfo.versions:
				try:
					nodeinfo_url = wellknown.url(ni_version)

				except KeyError:
					pass

			if not nodeinfo_url:
				raise ValueError('No valid nodeinfo urls')

		## Fetch and return nodeinfo
		with self.request(nodeinfo_url, raise_error=True) as resp:
			data = resp.text

			if not data:
				raise ValueError('Failed to fetch nodeinfo')

			return activitypub.Nodeinfo(data)


	def fetch_webfinger(self, handle, domain, instance=None):
		'Query webfinger for user urls'

		headers = {'Accept': 'application/json'}
		url = f'https://{domain}/.well-known/webfinger?resource=acct:{handle}@{instance or domain}'

		with self.request(url, headers=headers) as resp:
			if resp.status != 200:
				raise HttpClientError(resp)

			acct = resp.text

			if not acct:
				raise ValueError('Failed to fetch webfinger')

			return activitypub.Webfinger(acct)


	def fetch_object(self, url):
		'Fetch an ActivityPub object'

		headers = {'Accept': 'application/activity+json'}

		with self.request(url, headers=headers) as resp:
			if resp.status != 200:
				raise HttpClientError(resp)

			ap_object = resp.dict

			if not ap_object:
				raise ValueError('Failed to fetch ActivityPub object')

			return activitypub.Object.parse(ap_object)


class ClientHeaders(DotDict):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)


	def __getitem__(self, key):
		return super().__getitem__(key.replace('_', '-').title())


	def __setitem__(self, key, value):
		super().__setitem__(key.replace('_', '-').title(), value)


	def __delitem__(self, key):
		super().__delitem__(key.replace('_', '-').title())


	def update(self, data):
		for key, value in data.items():
			self[key] = value
