from ..datestring import DateString
from ..dotdict import CapitalDotDict
from ..enums import DatetimeFormat
from ..http_utils import CookieItem, Headers, create_request_message, methods
from ..misc import convert_to_bytes
from ..object_base import ObjectBase
from ..url import Url


class HttpClientRequest(ObjectBase):
	def __init__(self, url, method='GET', body=b'', headers={}, cookies={}, follow_redir=False):
		assert method.upper() in methods, f'Invalid method: {method}'

		super().__init__(
			headers = CapitalDotDict(headers),
			url = url,
			method = method,
			body = body,
			follow_redir = follow_redir,
			readonly_props = ['headers']
		)

		if 'Date' not in self.headers:
			self.headers['Date'] = DateString.now(DatetimeFormat.HTTP)


	def __bytes__(self):
		return self.compile()


	def __getitem__(self, key):
		return self.headers[key]


	def __setitem__(self, key, value):
		self.headers[key] = value


	def __delitem__(self, key):
		del self.headers[key]


	def _parse_property(self, key, value):
		if key == 'url':
			if not isinstance(value, Url):
				value = Url(value)
				self.headers['Host'] = value.host

			value = value.replace_property('anchor', None)

		elif key == 'body':
			if not isinstance(value, bytes):
				value = convert_to_bytes(value)

			self.headers['Content-Length'] = len(value)

		elif key == 'headers':
			value = Headers(value)

		return value


	@property
	def cookies(self):
		for cookie in self.headers.get('cookie', []):
			yield cookie


	def compile(self):
		if 'Content-Length' not in self.headers and self.body:
			self.headers['Content-Length'] = len(self.body)

		if 'Date' not in self.headers:
			self.headers['Date'] = DateString.now(DatetimeFormat.HTTP)

		return create_request_message(self.method, self.url.path, body=self.body, headers=self.headers)


	def get_cookie(self, key):
		return self.headers['Cookies']


	def get_header(self, key, default=None):
		try:
			return self.headers[key]

		except KeyError:
			self.headers[key] = default

		return self.headers[key]


	def set_header(self, key, value):
		self.headers[key] = value


	def set_cookie(self, key, value, **kwargs):
		for cookie in self.cookies:
			if key == cookie.key:
				cookie.value = value
				cookie.update(kwargs)
				return cookie

		item = CookieItem(key, value, **kwargs)
		self.cookies.append(item)
		return item
