from io import BytesIO, IOBase

from ..dotdict import DotDict
from ..http_utils import Headers
from ..url import Url


class HttpClientResponse(BytesIO):
	_response = None
	_connection = None

	def __init__(self, connection):
		self._response = connection.getresponse()
		self._connection = connection

		headers = {k:v for k,v in self._response.getheaders()}

		self.headers = Headers(headers, request=False, readonly=True)
		self._origin = None
		self._redirs = {}


	def __del__(self):
		self.close()


	def __enter__(self):
		return self


	def __exit__(self, *args):
		self.close()


	@property
	def origin(self) -> Url:
		return Url(self._origin or self._response.geturl())


	@property
	def reason(self) -> str:
		return self._response.reason


	@property
	def redir(self) -> Url:
		try:
			return self._redirs[-1]

		except IndexError:
			return None


	@property
	def redirs(self) -> list[Url]:
		return self._redirs


	@property
	def status(self) -> int:
		return self._response.status


	@property
	def url(self) -> Url:
		return self.redir or self.origin


	@property
	def version(self) -> float:
		return self._response.version / 10


	@property
	def length(self) -> int:
		try:
			data = self.headers['Content-Length'].one()

			if not data or data < 0:
				raise ValueError('Empty or negative value not allowed')

			return data

		except (KeyError, ValueError, IndexError):
			return 0


	@property
	def location(self) -> str:
		try:
			return self.headers['Location'].one()

		except (KeyError, IndexError):
			return None


	@property
	def body(self) -> IOBase:
		if not self._response.closed:
			self.seek(0,2)
			self.write(self._response.read())
			self.close()

		self.seek(0, 0)
		return self


	@property
	def bytes(self):
		return self.body.read()


	@property
	def text(self) -> str:
		return self.bytes.decode('utf-8')


	@property
	def dict(self) -> DotDict:
		return DotDict(self.bytes)


	def close(self):
		for buff in {self, self._response}:
			if not buff.closed:
				IOBase.close(buff)

		self._connection.close()


	def save(self, path, text=False):
		path = Path(path)
		mode = 'w' if text else 'wb'
		data = self.text if text else self.bytes

		if not path.parent.exists():
			path.parent.mkdir()

		with path.open(mode) as fd:
			fd.write(data)

		return path
