import json
import threading
import traceback

from functools import partial
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from http_router import Router, MethodNotAllowed, NotFound
from socketserver import ThreadingMixIn

from .request import Request
from .response import Response

from .. import __version__, exceptions, izzylog
from ..datestring import DateString
from ..dotdict import DotDict
from ..enums import DatetimeFormat
from ..http_utils import methods as http_methods, status_messages
from ..misc import convert_to_bytes, signal_handler
from ..object_base import ObjectBase
from ..path import Path
from ..url import Address


def get_headers(headers):
	return {key.title(): value for key, value in headers.items()}


class Application(ObjectBase):
	def __init__(self, listen='0.0.0.0', port=8080, host=None, headers={}, timeout=5, access_log=True):
		super().__init__(
			listen = listen,
			port = int(port),
			host = host or f'{listen}:{port}',
			headers = get_headers(headers),
			single_handler = None,
			access_log = access_log,
			timeout = timeout,
			router = Router(trim_last_slash=True),
			views = [],
			middleware = DotDict(pre=[], post=[]),
			ctx = DotDict()
		)

		self._server = None


	def __getattr__(self, key):
		try:
			return ObjectBase.__getattr__(self, 'ctx')[key]

		except (KeyError, TypeError, AttributeError):
			return ObjectBase.__getattr__(self, key)


	def __delattr__(self, key):
		try:
			del self.ctx[key]

		except KeyError:
			ObjectBase.__delattr__(self, key)


	def __getitem__(self, key):
		return self.ctx[key]


	def __setitem__(self, key, value):
		self.ctx[key] = value


	def __delitem__(self, key):
		del self.ctx[key]


	@property
	def running(self):
		return True if self._server else False


	def get_route(self, path, method='GET'):
		if self.single_handler:
			return ObjectBase(
				path = path,
				method = method,
				target = self.single_handler,
				params = {},
				readonly_props = True
			)

		match = self.router(str(path), method.upper())

		return ObjectBase(
			path = match.path,
			method = match.method,
			target = match.target,
			params = match.params or {},
			readonly_props = True
		)


	def add_route(self, handler, path, methods=['GET']):
		assert not self.single_handler, 'Server in single-handler mode'

		if not isinstance(methods, list):
			methods = [methods]

		for method in methods:
			if method.upper() not in http_methods:
				raise ValueError(f'Invalid HTTP method: {method}')

			self.router.bind(handler, str(path), methods=[method])


	def route(self, *paths, methods=['GET']):
		assert len(paths) > 0, 'Failed to specify paths for route'

		def wrapper(handler):
			for path in paths:
				self.add_route(handler, path, methods)

			return handler

		return wrapper


	def add_view(self, new_view, *args, **kwargs):
		assert len(new_view.paths) > 0, 'At least one path has to be specified for a view'

		view = new_view(self, *args, **kwargs)

		for method in http_methods:
			try:
				handler = view.get_handler(method)

			except AttributeError:
				continue

			for path in view.paths:
				self.add_route(handler, path, method)


	def view(self, *args, **kwargs):
		def wrapper(handler):
			self.add_view(handler, *args, **kwargs)

			return handler

		return wrapper


	def run(self):
		if self._server:
			return

		try:
			izzylog.info(f'Starting HTTP server @ {self.listen}:{self.port}')

			## Create server and set properties
			self._server = ThreadingHTTPServer((self.listen, self.port), partial(ClientHandler, self))
			self._server.allow_reuse_address = True
			self._server.timeout = self.timeout

			## Set signal handler if main thread
			if threading.current_thread() == threading.main_thread():
				signal_handler(run_func_in_thread, self._server.shutdown)
				izzylog.debug('Set signal handler')

			## Start server
			self._server.serve_forever()

		except Exception as e:
			raise e from None

		finally:
			self._server = None

		## Unset signal handler if main thread
		if threading.current_thread() == threading.main_thread():
			signal_handler(None)

		izzylog.info(f'Server shutdown')


	def set_single_handler(self, handler):
		self.single_handler = handler


	def unset_single_handler(self):
		self.single_handler = None


	def start(self):
		if self._server:
			return

		thread = threading.Thread(target=self.run)
		thread.start()

		signal_handler(self.stop)


	def stop(self):
		if not self._server:
			return

		self._server.shutdown()
		signal_handler(None)


class ClientHandler(BaseHTTPRequestHandler):
	def __init__(self, app, *args, **kwargs):
		self.app = app

		super().__init__(*args, **kwargs)


	def __getattr__(self, key):
		if key.startswith('do_'):
			method = key[3:].upper()

			if method not in http_methods:
				raise exceptions.HttpMethodNotAllowed(method)

			return object.__getattribute__(self, 'handle_request')

		return object.__getattribute__(self, key)


	@property
	def method(self):
		return self.command.upper()


	def read(self, length=None):
		return self.rfile.read(length)


	def write(self, data):
		if not isinstance(data, bytes):
			data = convert_to_bytes(data)

		self.wfile.write(data)

		return data.endswith(b'\r\n\r\n')


	def log_request(self, code, size):
		remote = self.headers.get('X-Real-Ip', self.headers.get('X-Forwarded-For', self.client_address[0]))
		path = self.path.split('?')[0]
		agent = self.headers.get('User-Agent', 'no agent')

		izzylog.info(f'{remote} {self.method} {path} {code} {size} {agent}')


	def send_headers(self, status, headers):
		## Combine default and new headers
		response_headers = self.app.headers.copy()
		response_headers.update({key.title(): value for key, value in headers.items()})

		## Make sure the default headers are set
		if 'Content-Type' not in response_headers:
			response_headers['Content-Type'] = 'text/plain; charset=utf-8'

		if 'charset' not in response_headers['Content-Type']:
			response_headers['Content-Type'] += '; charset=utf-8'

		if 'Server' not in response_headers:
			response_headers['Server'] = f'IzzyLib/{__version__}'

		if 'Date' not in response_headers:
			response_headers['Date'] = DateString.now(DatetimeFormat.HTTP)

		## Send the whole response
		self.send_response_only(status, status_messages[status])

		for key, value in response_headers.items():
			self.send_header(key, value)

		self.end_headers()


	def send_body(self, body):
		if not self.write(body):
			self.write(b'\r\n\r\n')


	def send_error(self, message, status=500):
		html = f'''<html>
	<head>
		<title>HTTP Error {status}</title>
		<style>
			body {{
				color: #eee;
				background: #111;
			}}
			a {{
				color: #d29;
			}}
			a:hover {{
				color: #d6b
			}}
		</style>
	</head>
	<body>
		<div style="text-align:center;margin:25px 0;">HTTP Error {status}: {message}</div>
		<hr>
		<a href="https://git.barkshark.xyz/izaliamae/izzylib" target="_blank">IzzyLib HTTP Server/{__version__}</a>
	</body>
</html>'''

		self.send_response(html, status, {'Content-Type': 'text/html'})


	def send_response(self, body=b'', status=200, headers={}):
		self.send_headers(status, headers)
		self.send_body(body)

		## Log the request
		if self.app.access_log:
			self.log_request(status, headers.get('Content-Length'))


	def handle_request(self):
		try:
			request = Request(self)
			match = self.app.get_route(request.path, self.method)
			response = match.target(request, **match.params)

			if not isinstance(response, Response):
				logging.error(f'Invalid handler response for {request.path}: {type(response).__name__}')
				raise exceptions.HttpServerError('Invalid handler response')

			self.send_response(response.body, response.status, response.headers)

		except NotFound:
			self.send_error(f'Not Found', 404)

		except MethodNotAllowed:
			self.send_error(f'Method Not Allowed ({self.method})', 405)

		except exceptions.HttpError as e:
			self.send_error(e.message, e.status)

		except Exception as e:
			traceback.print_exc()
			self.send_error('Server Error', 500)


def run_func_in_thread(func, *args, **kwargs):
	thread = threading.Thread(target=func, args=args, kwargs=kwargs)
	thread.start()
