import asyncio
import json
import re

from datetime import datetime, timezone
from io import BytesIO

from .datestring import DateString
from .dotdict import DotDict, ReadOnlyDotDict
from .enums import HttpMethod
from .exceptions import RouteFoundError, RouteNotFoundError
from .misc import Boolean, class_name
from .object_base import ObjectBase
from .path import Path
from .url import Address, Url

from typing import Dict


methods = [
	'CONNECT',
	'DELETE',
	'GET',
	'HEAD',
	'OPTIONS',
	'PATCH',
	'POST',
	'PUT',
	'TRACE'
]


status_messages = {
	100: 'Continue',
	101: 'Switching Protocols',
	103: 'Early Hints',
	200: 'Ok',
	201: 'Created',
	202: 'Accepted',
	203: 'Non-Authoritative Information',
	204: 'No Content',
	205: 'Reset Content',
	206: 'Partial Content',
	300: 'Multiple Choice',
	301: 'Moved Permanently',
	302: 'Found',
	303: 'See Other',
	304: 'Not Modified',
	307: 'Temporary Redirect',
	308: 'Permanent Redirect',
	400: 'Bad Request',
	401: 'Unauthorized',
	402: 'Payment Required',
	403: 'Forbidden',
	404: 'Not Found',
	405: 'Method Not Allowed',
	406: 'Not Acceptable',
	407: 'Proxy Authentication Required',
	408: 'Request Timeout',
	409: 'Conflit',
	410: 'Gone',
	411: 'Length Required',
	412: 'Precondition Failed',
	413: 'Payload Too Large',
	414: 'URI Too Long',
	415: 'Usupported Media Type',
	416: 'Range Not Satisfiable',
	417: 'Expectation Failed',
	418: 'I\'m A Teapot',
	421: 'Misdirected Request',
	422: 'Unprocessable Entity',
	425: 'Too Early',
	426: 'Upgrade Required',
	428: 'Precondition Required',
	429: 'Too Many Requests',
	431: 'Request Header Fields Too Large',
	451: 'Unavailable For Legal Reasons',
	500: 'Internal Server Error',
	501: 'Not Implemented',
	502: 'Bad Gateway',
	503: 'Service Unavailable',
	504: 'Gateway Timeout',
	505: 'HTTP Version Not Supported',
	506: 'Variant Also Negotiates',
	507: 'Not Extended',
	511: 'Network Authentication Required',

	# Webdav statuses
	102: 'Processing',
	207: 'Multi-Status',
	208: 'Already Reported',
	226: 'IM Used',
	423: 'Locked',
	424: 'Failed Dependency',
	507: 'Insufficient Storage',
	508: 'Loop Detected'
}


class Headers(DotDict):
	_readonly = False
	_cookie_header = None


	def __init__(self, *args, request=True, readonly=False, **kwargs):
		super().__init__(*args, **kwargs)
		self._cookie_header = 'Cookie' if request else 'Set-Cookie'
		self._readonly = readonly


	def __getitem__(self, key):
		return super().__getitem__(self._format_key(key))


	def __setitem__(self, key, value):
		if self._readonly:
			raise ReadOnlyError('Headers are readonly')

		key = self._format_key(key)

		if type(value) == HeaderItem:
			super().__setitem__(key, value)
			return

		if key == self._cookie_header:
			if type(value) == str:
				value = CookieItem.from_string(value)

			elif type(value) != CookieItem:
				raise TypeError(f'{key} header must be a `str` or `CookieItem`, not a `{class_name(value)}`')

			try:
				for cookie in self[key]:
					if cookie.key == value.key:
						new_data = {k:v for k,v in value.as_dict().items() if k != key}

						cookie.value = value.value
						cookie.update()
						return

				self[key].append(value)
				return

			except KeyError:
				pass

		elif key == 'Date':
			value = DateString(value, 'http')

		elif key == 'Content-Length':
			try:
				value = int(value)

			except ValueError:
				return

		super().__setitem__(key, HeaderItem(key, value))


	def __delitem__(self, key):
		super().__delitem__(self._format_key(key))


	def _format_key(self, key):
		return key.replace('_', '-').title()


	def append(self, key, value):
		try:
			self[key].append(value)

		except KeyError:
			self[key] = value


	def cookies(self):
		for cookie in self.get(self._cookie_header, []):
			yield cookie


	def items(self):
		for key, item in super().items():
			for value in item:
				yield (key, value)


	def get(self, key, default=None):
		try:
			self.__getitem__(key)

		except KeyError:
			return default


	def get_cookie(self, key):
		for cookie in self.cookies():
			if cookie.key == key:
				return cookie

		raise KeyError(key)


	def get_one(self, key, default=None):
		try:
			return self[key].one()

		except (KeyError, IndexError):
			return default


	def new_cookie(self, key, value, **kwargs):
		return self.set_cookie(CookieItem(key, value, **kwargs))


	def new_cookie_from_string(self, value):
		return self.set_cookie(CookieItem.from_string(value))


	def set_all(self, key, value):
		try:
			self[key].set(value)

		except KeyError:
			self[key] = value


	def set_cookie(self, cookie):
		self[self._cookie_header] = cookie
		return cookie


class HeaderItem(list):
	def __init__(self, key, values):
		super().__init__()
		self.update(values)

		self.key = key


	def __str__(self):
		return ','.join(str(v) for v in self)


	def __repr__(self):
		return f'HeaderItem({self.key})'


	def set(self, *values):
		self.clear()

		for value in values:
			self.append(value)


	def one(self):
		return self[0]


	def update(self, *items):
		for item in items:
			self.append(item)


class CookieItem(DotDict):
	def __init__(self, key, value, **kwargs):
		self.key = key
		self.value = value
		self.args = DotDict()

		for k,v in kwargs.items():
			if k not in cookie_params.values():
				raise AttributeError(f'Not a valid cookie parameter: {key}')

			setattr(self, k, v)


	def __str__(self):
		text = f'{self.key}={self.value}'

		if self.expires:
			text += f'; Expires={self.expires.strftime("%a, %d %b %Y %H:%M:%S GMT")}'

		if self.maxage != None:
			text += f'; Max-Age={self.maxage}'

		if self.domain:
			text += f'; Domain={self.domain}'

		if self.path:
			text += f'; Path={self.path}'

		if self.samesite:
			text += f'; SameSite={self.samesite}'

		if self.secure:
			text += f'; Secure'

		if self.httponly:
			text += f'; HttpOnly'

		return text


	@classmethod
	def from_string(cls, data):
		kwargs = {}

		for idx, pairs in enumerate(data.split(';')):
			try:
				k, v = pairs.split('=', 1)
				v = v.strip()

			except:
				k, v = pairs, True

			k = k.replace(' ', '')

			if isinstance(v, str) and v.startswith('"') and v.endswith('"'):
				v = v[1:-1]

			if idx == 0:
				key = k
				value = v

			elif k in cookie_params.keys():
				kwargs[cookie_params[k]] = v

			else:
				logging.info(f'Invalid key/value pair for cookie: {k} = {v}')

		return cls(key, value, **kwargs)


	@property
	def expires(self):
		return self.args.get('Expires')


	@expires.setter
	def expires(self, data):
		if isinstance(data, str):
			data = DateString(data, 'http')

		elif isinstance(data, int) or isinstance(data, float):
			data = datetime.fromtimestamp(data).replace(tzinfo=timezone.utc)

		elif isinstance(data, datetime):
			if not data.tzinfo:
				data = data.replace(tzinfo=timezone.utc)

		elif isinstance(data, timedelta):
			data = datetime.now(timezone.utc) + data

		else:
			raise TypeError(f'Expires must be a http date string, timestamp, or datetime object, not {data.__class__.__name__}')

		self.args['Expires'] = data


	@property
	def maxage(self):
		return self.args.get('Max-Age')


	@maxage.setter
	def maxage(self, data):
		if isinstance(data, int):
			pass

		elif isinstance(date, timedelta):
			data = data.seconds

		elif isinstance(date, datetime):
			data = (datetime.now() - date).seconds

		else:
			raise TypeError(f'Max-Age must be an integer, timedelta object, or datetime object, not {data.__class__.__name__}')

		self.args['Max-Age'] = data


	@property
	def domain(self):
		return self.args.get('Domain')


	@domain.setter
	def domain(self, data):
		if not isinstance(data, str):
			raise ValueError(f'Domain must be a string, not {data.__class__.__name__}')

		self.args['Domain'] = data


	@property
	def path(self):
		return self.args.get('Path')


	@path.setter
	def path(self, data):
		if not isinstance(data, str):
			raise ValueError(f'Path must be a string or izzylib.Path object, not {data.__class__.__name__}')

		self.args['Path'] = Path(data)


	@property
	def secure(self):
		return self.args.get('Secure')


	@secure.setter
	def secure(self, data):
		self.args['Secure'] = boolean(data)


	@property
	def httponly(self):
		return self.args.get('HttpOnly')


	@httponly.setter
	def httponly(self, data):
		self.args['HttpOnly'] = boolean(data)


	@property
	def samesite(self):
		return self.args.get('SameSite')


	@samesite.setter
	def samesite(self, data):
		if isinstance(data, bool):
			data = 'Strict' if data else 'None'

		elif isinstance(data, str) and data.title() in ['Strict', 'Lax', 'None']:
			data = data.title()

		else:
			raise TypeError(f'SameSite must be a boolean or one of Strict, Lax, or None, not {data.__class__.__name__}')

		self.args['SameSite'] = data
		self.args['Secure'] = True


	def as_dict(self):
		data = DotDict({self.key: self.value})
		data.update(self.args)

		return data


	def set_defaults(self):
		for key in list(self.args.keys()):
			del self.args[key]


	def set_delete(self):
		self.args.pop('Expires', None)
		self.maxage = 0

		return self


class Router(ObjectBase):
	def __init__(self, regex=None, **types):
		ObjectBase.__init__(self,
			routes = DotDict({method.value: DotDict() for method in HttpMethod}),
			trailing_slash = True,
			regex = re.compile(regex or r'\{([A-Za-z0-9_:]+)\}'),
			types = DotDict(),
			readonly_props = ['routes', 'types']
		)

		self.types.update(
			str = str,
			bool = Boolean,
			int = int,
			float = float,
			ts = datetime.fromtimestamp,
			addr = Address
		)

		self.types.update(types)


	def parse_path(self, path):
		if not path.startswith('/'):
			path = '/' + path

		if not path.endswith('/') and self.trailing_slash:
			path += '/'

		return path


	def add_route(self, handler, path, method=HttpMethod.GET):
		if not isinstance(method, HttpMethod):
			method = HttpMethod[method.upper()]

		path = self.parse_path(path)

		if path in self.routes[method.value]:
			raise KeyError(f'Handler for path already exists: {method.value} {path}')

		parts = []

		for idx, part in enumerate(path.split('/')):
			path_parts = self.regex.findall(part)
			name = None
			type = 'str'
			pre = ''
			post = ''

			if len(path_parts) > 1:
				raise RuntimeError('Only one variable per path part is supported')

			elif not len(path_parts):
				parts.append(DotDict(
					part = part,
					name = name,
					type = type,
					pre = pre,
					post = post
				))

			else:
				varname = '{' + path_parts[0] + '}'
				parsed = path_parts[0].split(':')

				if len(parsed) == 2:
					name = parsed[0]
					type = parsed[1]

				elif len(parsed) == 1:
					name = parsed[0]

				index = part.find(varname)

				if index == 0:
					post = part.split(varname)[0]

				else:
					try:
						pre, post = part.split(varname)

					except IndexError:
						pre = part.split(varname)[0]

				parts.append(DotDict(
					part = part,
					name = name,
					type = type,
					pre = pre,
					post = post
				))

		route = RouterRoute(self, method, path, handler, parts)
		self.routes[method.value][path] = route

		return route


	def del_route(self, path, method='GET'):
		if not isinstance(method, HttpMethod):
			method = HttpMethod[method.upper()]

		try:
			del self.routes[method.value][path]

		except KeyError:
			raise RouteFoundError(path, method) from None

	def match(self, path, method='GET'):
		if not isinstance(method, HttpMethod):
			method = HttpMethod[method.upper()]

		for route in self.routes[method.value].values():
			match = route.match(path)

			if match != False:
				return match

		raise RouteNotFoundError(path, method)


	def route(self, *paths):
		def wrapper(func):
			for path in paths:
				self.add_route(func, path)

			return func
		return wrapper


class RouterRoute(ObjectBase):
	def __init__(self, router, method, path, handler, parts):
		ObjectBase.__init__(self,
			router = router,
			method = method,
			path = path,
			handler = handler,
			parts = parts,
			readonly_props = True
		)


	def __repr__(self):
		return f'RouterRoute("{self.method.value} {self.path}")'


	@property
	def length(self):
		return len(self.parts)


	def match(self, path):
		if path == self.path:
			return RouterMatch(self.router, self.method, path, self.handler)

		split = self.router.parse_path(path).split('/')

		if self.length != len(split):
			return False

		kwargs = {}

		for idx, part in enumerate(split):
			part_match = self.parts[idx]

			if part_match.part == part:
				continue

			elif not part_match.name:
				return False

			value = part.replace(part_match.pre, '', 1).replace(part_match.post, '', -1)

			try:
				kwargs[part_match.name] = self.router.types[part_match.type](value)

			except Exception as e:
				print(class_name(e), e)
				return False

		return RouterMatch(self.router, self.method, path, self.handler, **kwargs)


class RouterMatch(ReadOnlyDotDict):
	def __init__(self, router, method, path, handler, **kwargs):
		DotDict.__init__(self, kwargs)

		self._router = router
		self._method = method
		self._path = path
		self._handler = handler


	def __repr__(self):
		return f'RouterMatch("{self.method.value} {self.path}", handler="{class_name(self.handler)}")'


	def __str__(self):
		return self.__repr__()


	@property
	def router(self):
		return self._router


	@property
	def path(self):
		return self._path


	@property
	def handler(self):
		return self._handler


	@property
	def method(self):
		return self._method


	def run(self, *args, **kwargs):
		return self.handler(*args, **kwargs, **self)


def parse_headers(raw_headers, request=True):
	data = DotDict(
		headers = []
	)

	for idx, line in enumerate(raw_headers.decode('utf-8').splitlines()):
		if idx == 0:
			if request:
				method, path, version = line.split()

				data.update({
					'method': method,
					'path': Path(path)
				})

			else:
				version, status, *message = line.split()

				data.update({
					'status': int(status),
					'message': ' '.join(message)
				})

		else:
			try: key, value = line.split(': ', 1)
			except: continue

			data.headers.append((key, value))

	data.version = float(version.split('/')[1])

	return data


def create_message(first_line: str, headers: Dict[str,str], body: bytes=None):
	data = first_line

	for k,v in headers.items():
		data += f'\r\n{k}: {v}'

	data += '\r\n\r\n'
	data = data.encode('utf-8')

	if body:
		if not body.endswith(b'\r\n\r\n'):
			body += b'\r\n\r\n'

		return data + body

	return data


def create_request_message(method, path, body=None, headers=None, version='1.1'):
	if not isinstance (method, HttpMethod):
		method = HttpMethod[method.upper()]

	return create_message(f'{method.value} {path} HTTP/{version}', headers, body)


def create_response_message(status, message=None, body=None, headers=None, version='1.1'):
	msg = f'HTTP/{version} {status}'

	if message:
		msg += f' {message}'

	return create_message(msg, headers, body)
