import os, queue, sys, threading, time

from datetime import datetime
from pathlib import Path

from .enums import LogLevel


class Log:
	__slots__ = [
		'name', 'level', 'date', 'format', 'logfile',
		'lockfile', 'critical', 'error',
		'warning', 'info', 'verbose', 'debug', 'merp'
	]

	def __init__(self, name, **config):
		self.name = name
		self.level = LogLevel.INFO
		self.date = True
		self.format = '%Y-%m-%d %H:%M:%S'
		self.logfile = None

		self.update_config(**config)

		for level in LogLevel:
			self._set_log_function(level)


	def _set_log_function(self, level):
		setattr(self, level.name.lower(), lambda *args: self.log(level, *args))


	def print(self, *args):
		sys.stdout.write(' '.join([str(arg) for arg in args]))
		sys.stdout.flush()


	def parse_level(self, level):
		try: return LogLevel(int(level))
		except ValueError: return LogLevel[level.upper()]


	def set_level_from_env(self, env_name):
		level = os.environ.get(env_name)

		if not level:
			self.verbose('Invalid environmental variable:', env_name)
			return

		self.set_config('level', level)


	def update_config(self, **data):
		for key, value in data.items():
			self.set_config(key, value)


	def set_config(self, key, value):
		if key == 'level' and type(value) in [str, int]:
			value = self.parse_level(value)

		if key == 'logfile':
			value = LogFile(value)

		setattr(self, key, value)


	def get_config(self, key):
		return getattr(self, key)


	def print_config(self):
		self.print(*(f'{k}: {v}\n' for k,v in self.items()))


	def log(self, level, *msg):
		if isinstance(level, str):
			LogLevel[level.upper()]

		if level < self.level:
			return

		default = self.name == 'Default'
		options = [
			level.name + ':',
			' '.join([str(message) for message in msg]),
			'\n'
		]

		if self.date and not os.getppid() == 1:
			options.insert(0, datetime.now().strftime(self.format))

		if not default:
			options.insert(0 if not self.date else 1, f'[{self.name}]')

		if self.logfile:
			self.logfile.write(*options)

		self.print(*options)


class LogFile:
	def __init__(self, path):
		super().__init__()

		self.daemon = True
		self.path = Path(path).resolve()
		self.fd = None
		self.lockfile = Path(str(self.path)+'.lck')
		self.lockfd = None

		## Make sure a lock file does not already exist
		self.is_locked
		self.lock = threading.Lock()


	def __enter__(self):
		self.aquire_file_lock()

		if not self.fd:
			self.fd = self.path.open('a')

		return self


	def __exit__(self):
		self.release_file_lock()


	def __del__(self):
		self.close()
		self.release_file_lock()


	@property
	def is_locked(self):
		if not self.lockfd and self.lockfile.exists():
			with self.lockfile.open('r') as fd:
				try: pid = int(fd.read())
				except: return False

			if self.pid != pid and check_pid(pid):
				raise FileExistsError('Cannot aquire lock')

			self.lockfile.unlink()
			return True

		if self.lockfd:
			return True

		return False


	@property
	def pid(self):
		return os.getpid()


	def aquire_file_lock(self):
		if self.is_locked:
			return

		self.lockfd = self.lockfile.open('w')
		self.lockfd.write(str(self.pid))
		self.lockfd.flush()


	def release_file_lock(self):
		try:
			if not self.is_locked:
				return

		except FileExistsError:
			return

		self.lockfd.close()
		self.lockfile.unlink()
		self.lockfd = None


	def open(self):
		if self.fd:
			return

		self.fd = self.path.open('a')


	def close(self):
		if not self.fd:
			return

		self.fd.close()
		self.fd = None


	def write(self, *message):
		if not self.is_locked:
			self.aquire_file_lock()

		if not self.fd:
			self.open()

		with self.lock:
			self.fd.write(' '.join(message))
			self.fd.flush()


def get_logger(name, **config):
	try:
		return logger[name.lower()]

	except KeyError:
		log = Log(name, **config)
		logger[name.lower()] = log
		return log


def check_pid(pid: int) -> bool:
	'Check if a process with `PID` is running'

	assert isinstance(pid, int)

	try:
		os.kill(pid, 0)
		return True

	except OSError:
		return False


'''create a default logger'''
logger = {
	'default': Log('Default'),
	'IzzyLib': Log('IzzyLib')
}

DefaultLog = logger['default']

'''aliases for default logger's log output functions'''
critical = DefaultLog.critical
error = DefaultLog.error
warning = DefaultLog.warning
info = DefaultLog.info
verbose = DefaultLog.verbose
debug = DefaultLog.debug
merp = DefaultLog.merp
log = DefaultLog.log

'''aliases for the default logger's config functions'''
set_level_from_env = DefaultLog.set_level_from_env
update_config = DefaultLog.update_config
set_config = DefaultLog.set_config
get_config = DefaultLog.get_config
print_config = DefaultLog.print_config


logger['IzzyLib'].set_config('level', os.environ.get('IZZYLIB_LOG_LEVEL', 'INFO'))
