import json, re, select

from . import __version__
from .config import BaseConfig
from .datestring import DateString
from .dotdict import DotDict
from .exceptions import HttpError, NoConnectionError, TokenExistsError, VersionError
from .http_client import HttpClient
from .misc import join_list
from .object_base import ObjectBase
from .url import Url


vpattern = re.compile('([0-9]+)\.([0-9]+)\.([0-9]+)')


def version_check(minimum=None, maximum=None):
	min_tpl = tuple(minimum.split('.')) if minimum else None
	max_tpl = tuple(maximum.split('.')) if maximum else None

	def api_deco(func):
		def funcwrap(self, *args, **kwargs):
			if not self.version:
				raise NoConnectionError('Connect to the instance fist')

			if minimum and self.version < min_tpl:
				raise VersionError(f'Instance version must be at least {minimum}. Current version: {self.version_string}')

			elif maximum and self.version > max_tpl:
				raise VersionError(f'Endpoint does not exist anymore')

			return func(self, *args, **kwargs)

		return funcwrap
	return api_deco


class MastodonApi(ObjectBase):
	def __init__(self, domain, token=None, appname='Mastodon Api Client', timeout=5, **kwargs):
		super().__init__(
			appname = appname,
			domain = domain,
			client_id = kwargs.get('client_id'),
			client_secret = kwargs.get('client_secret'),
			token = token,
			vapid_key = kwargs.get('vapid_key'),
			redirect = kwargs.get('redirect', 'urn:ietf:wg:oauth:2.0:oob'),
			scopes = Scopes(*kwargs.get('scopes', [])),
			website = kwargs.get('website'),
			version = None,
			readonly = ['appname', 'domain', 'website']
		)

		self._client = HttpClient(
			appagent = appname,
			timeout = timeout,
			headers = {
				'Accept': 'application/json',
				'Accept-Encoding': 'identity'
			}
		)


	def __enter__(self):
		self.connect()
		return self


	def __exit__(self, *args):
		self.disconnect()


	@property
	def version_string(self):
		return join_list(self.version)


	def connect(self):
		instance = self.api('/api/v1/instance')

		try:
			version = vpattern.findall(instance['version'])[0]
		except IndexError:
			raise ConnectionError('Unable to find instance version')

		self.version = version


	def disconnect(self):
		self.version = None


	def api(self, endpoint, data={}, token=None, method='GET'):
		if method.upper() not in ['GET', 'POST', 'PUT', 'PATCH', 'DELETE']:
			raise ValueError(f'Invalid HTTP method: {method}')

		kwargs = DotDict(
			headers = {}
		)

		url = Url.new(self.domain, path=endpoint, query=data, proto='HTTPS')

		if data:
			if method == 'GET':
				url = url.replace_property('query', data)

			else:
				kwargs.body = data

		if token or self.token:
			kwargs.headers.authorization = f'Bearer {token or self.token}'

		if method.upper() in ['POST']:
			kwargs.headers.headers.content_type = 'application/json'

		with self._client.request(url, method, **kwargs, raise_error=True) as response:
			body_lines = response.text.splitlines()
			body = '\n'.join(body_lines[1:-2]) if len(body_lines) > 1 else body_lines[0]

			return json.loads(body)


	@version_check()
	def create_token(self, type, id, secret, auth_code=None, username=None, password=None):
		if self.token:
			raise TokenExistsError('Refusing to create a new token')

		data = DotDict(
			grant_type = type,
			client_id = id,
			client_secret = secret,
			redirect_uri = self.redirect,
			scope = self.scopes.compile()
		)

		if auth_code:
			data.code = auth_code

		if username and password:
			data.username = username
			data.password = password

		return self.api('/oauth/token', data, method='POST')


	def login_client(self, id=None, secret=None):
		return self.create_token('client_credentials',
			id or self.client_id,
			secret or self.client_secret
		)


	## Does not work atm for some reason (returns 406 Not Acceptable)
	@version_check()
	def authorize(self, id, force_login=False):
		data = DotDict(
			response_type = 'code',
			client_id = id,
			redirect_uri = self.redirect,
			scope = self.scopes.compile(),
			force_login = force_login
		)

		return self.api('/oauth/authorize', data)


	@version_check()
	def revoke(self, client_id, client_secret, token=None):
		assert token or self.token

		data = dict(
			client_id = client_id,
			client_secret = client_secret,
			token = token or self.token
		)

		return self.api('/oauth/revoke', data, token=token, method='POST')


	@version_check()
	def create_app(self):
		if self.token:
			raise TokenExistsError('Refusing to create a new app')

		data = DotDict(
			client_name = self.appname,
			redirect_uris = self.redirect
		)

		if self.scopes:
			data.scopes = self.scopes.compile()

		if self.website:
			data.website = Url(self.website)

		app = self.api('/api/v1/apps', data=data, method='POST')

		self.client_id = app.client_id
		self.client_secret = app.client_secret

		return app


	@version_check('2.0.0')
	def verify_credentials(self):
		return self.api('/api/v1/apps/verify_credentials')


	@version_check('2.7.0')
	def register(self, username, email, password, locale='en-us', reason=None):
		app = self.create_app()
		data = DotDict(
			username = username,
			email = email,
			password = password,
			locale = locale,
			agreement = True
		)

		if reason:
			data.reason = reason

		token = self.login_client(app.client_id, app.client_secret)

		return self.api('/api/v1/accounts', data, token=token.access_token, method='POST')


	@version_check('1.1.1')
	def update_profile(self, kwargs):
		return self.api('/api/v1/accounts/update_credentials', kwargs, method='PATCH')


	@version_check()
	def me(self):
		return Account(self.api('/api/v1/accounts/verify_credentials'))


	@version_check('1.1.0')
	def instance(self):
		return self.api('/api/v1/instance')


class Scopes:
	def __init__(self, *scopes):
		self.__scopes = []

		for scope in scopes:
			self.add_string(scope)


	def __str__(self):
		return self.compile()


	def __iter__(self):
		for scope in self.__scopes:
			yield scope


	def __getitem__(self, key):
		return self.__scopes[key]


	def __delitem__(self, key):
		del self.__scopes[key]


	def __len__(self):
		return len(self.__scopes)


	def add(self, type, name=None, admin=False):
		if self.index(type, name, admin) != None or self.index(type, admin=admin) != None:
			return

		if not name:
			rem_scopes = []

			for scope in self:

				if scope.type == type and scope.admin == admin:
					rem_scopes.append(scope)

			for scope in rem_scopes:
				self.remove(scope)

		scope = Scope(type, name, admin)
		self.__scopes.append(scope)

		return scope


	def add_list(self, *scopes):
		for scope in scopes:
			self.add(*scope)


	def add_string(self, string):
		item = Scope(string)
		self.add(item.type, item.name, item.admin)


	def index(self, type, name=None, admin=False):
		for idx, scope in enumerate(self):
			if scope.type == type and scope.name == name and scope.admin == admin:
				return idx


	def delete(self, type, name=None, admin=False):
		index = self.index(type, name, admin)

		if index == None:
			return

		del self[index]


	def remove(self, item):
		self.__scopes.remove(item)


	def compile(self):
		return ' '.join(scope.compile() for scope in self)


class Scope:
	BASE = {'accounts', 'blocks', 'bookmarks', 'favourites', 'filters', 'follows', 'lists', 'mutes', 'notifications', 'statuses'}
	READ = {'search', *BASE}
	WRITE = {'conversations', 'media', *BASE}
	ADMIN = {'accounts', 'reports'}


	def __init__(self, type, name=None, admin=False):
		self.type = type
		self.name = name
		self.admin = bool(admin)


	def __str__(self):
		return f'Scope(type={self.type}, name={self.name}, admin={self.admin})'


	@classmethod
	def new_from_string(cls, string):
		split = string.split(':')

		if len(split) == 1:
			type = split[0]
			name = None
			admin = False

		elif len(split) == 2:
			type = split[0]
			name = split[1]
			admin = False

		elif len(split) == 3:
			type = split[1]
			name = split[2]
			admin = True

		else:
			raise TypeError(f'Invalid scope: {string}')

		if type.lower() not in ['read', 'write']:
			raise ValueError(f'Invalid scope type: {type}')

		if admin:
			if name not in self.ADMIN:
				raise ValueError(f'Invalid admin scope name: {name}')
		
		elif name and name not in getattr(self, name.upper()):
			raise ValueError(f'Invalid {type} scope name: {name}')

		return cls(type, name, admin)


	def compile(self):
		string = ''

		if self.admin:
			string += 'admin:'

		string += self.type

		if self.name:
			string += f':{self.name}'

		return string


class Account(DotDict):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.created_at = DateString.new_activitypub(self.created_at)
		self.last_status_at = DateString(self.last_status_at, 'mastodon')


	@property
	def created(self):
		return self.created_at


	@property
	def last_status(self):
		return self.last_status_at


	@property
	def info_table(self):
		return DotDict()


class InfoTable(list):
	def __init__(self, account, table):
		super().__init__(table)

		self._account = account
		self._table = DotDict()
		self._index = []

		for row in table:
			self[row.name] = self.row.value


	def __getitem__(self, key):
		return self._table[key]


	def __setitem__(self, key, value):
		self._table[key] = value

		try:
			self[self._index(key)]['value'] = value

		except IndexError:
			self._index.append(key)


	def __delitem__(self, key):
		super().__delitem__(key)
		self._index.remove(key)


	def __getattr__(self, key):
		return self.__getitem__(key)


	def __setattr__(self, key, value):
		self.__setitem__(key, value)


	def keys(self):
		return self._table.keys()


	def values(self):
		return self._table.values()


	def items(self):
		return self._table.items()


	def get(self, key, default=None):
		return self._table.get(key, default)


	def update(self, data):
		for key, value in data.items():
			self[key] = value


	def pop(self, key, default=None):
		try:
			value = self._table.pop(key)
			self._index.remove(key)
			return key

		except KeyError as e:
			if not default:
				raise e from None

			return default


	def index(self, key):
		return self._index.index(key)
