import argparse
import asyncio
import hashlib
import json
import os
import platform
import random
import shlex
import signal
import socket
import statistics
import string
import time
import timeit

from datetime import datetime, timedelta, timezone
from functools import cached_property, wraps
from getpass import getpass, getuser
from importlib import util
from subprocess import Popen, PIPE
from urllib.parse import urlparse, quote, quote_plus
from urllib.request import urlopen

from . import izzylog
from .dotdict import DotDict
from .enums import DatetimeFormat
from .path import Path

# typing modules
from collections.abc import Callable
from typing import Any, Dict, List, Union

try:
	import grp

except ImportError:
	grp = None


__all__ = [
	'app_data_dirs',
	'catch_kb_interrupt',
	'character_check',
	'check_pid',
	'class_name',
	'convert_to_boolean',
	'convert_to_bytes',
	'convert_to_string',
	'fuzzy_string_match',
	'get_current_user_info',
	'get_ip',
	'hasher',
	'import_from_path',
	'nfs_check',
	'port_check',
	'print_methods',
	'prompt',
	'random_gen',
	'random_int',
	'random_port',
	'random_str',
	'remove',
	'replace_strings',
	'signal_handler',
	'sudo',
	'time_function',
	'time_function_pprint',
	'timestamp',
	'var_name',
	'ArgParser',
	'Argument',
	'Boolean'
]


def app_data_dirs(author:str, name:str) -> Dict[str,Path]:
	'Returns the cache and config paths for software'

	if platform.system() == 'Linux':
		config = Path('~/.config')
		cache = Path('~/.cache')

	elif platform.system() == 'Darwin':
		config = Path('~/Application Support')
		cache = Path('~/Library/Caches')

	elif platform.system() == 'Windows':
		return DotDict(
			config = Path('~/Application Data/Local Settings').join(author, name),
			cache = Path('~/Application Data').join(author, name, 'Cache')
		)

	else:
		raise TypeError(f'Unknown system: {platform.system()}')

	return DotDict(
		config = config.join(author, name),
		cache = cache.join(author, name)
	)


def catch_kb_interrupt(function, *args, **kwargs):
	'Run a function and catch the KeyboardInterrupt exception'

	try:
		return function(*args, **kwargs)

	except KeyboardInterrupt as e:
		izzylog.verbose('Bye! UvU')
		return e


def character_check(string: str, allowed: str = string.ascii_lowercase + string.digits) -> bool:
	'Check if a string only contains certain characters'

	return set(string).issubset(allowed)


def check_pid(pid: int) -> bool:
	'Check if a process with `PID` is running'

	assert isinstance(pid, int)

	try:
		os.kill(pid, 0)
		return True

	except OSError:
		return False


def class_name(cls:object) -> str:
	'Get the name of a class'
	try:
		name = cls.__name__

	except:
		name = cls.__class__.__name__

	try:
		return name.split('.')[-1]

	except:
		return name


def convert_to_boolean(value:Union[str,bool,int,type(None)], return_value:bool=False) -> Union[bool,Any]:
	'Convert a str, bool, int or None object into a boolean.'

	try:
		return Boolean(value).raw()

	except TypeError as e:
		if return_value:
			return value

		raise e from None


def convert_to_bytes(data:Union[str,DotDict,dict,list,set,tuple,bytearray,bytes], encoding:str='utf-8') -> bytes:
	'Attempt to convert an object to bytes. Only supports str, `izzylib.dotdict.DotDict`, dict, list, set, tuple, and bytearray objects.'

	if not data:
		return b''

	if isinstance(data, bytearray):
		return bytes(data)

	elif not isinstance(data, bytes):
		return convert_to_string(data).encode(encoding)

	return data


def convert_to_string(data:Union[str,DotDict,dict,list,set,tuple,bytearray,bytes]) -> str:
	'Attempt to convert an object to a str.'

	if not data:
		return ''

	if isinstance(data, DotDict):
		return data.to_json()

	elif isinstance(data, (dict, list, set, tuple)):
		return json.dumps(data)

	elif isinstance(data, (bytes, bytearray)):
		return data.decode('utf-8')

	elif not isinstance(data, str):
		raise TypeError(f'Invalid data type: {class_name(data)}')

	return data


def fuzzy_string_match(string: str, *values: str, case_sensitive=False) -> List[bool]:
	'Check if a string is in a list of values. Not case-sensitive by default.'

	result = []

	for value in values:
		if not isinstance(value, str):
			result.append(False)

		else:
			if case_sensitive:
				result.append(string in value)

			else:
				result.append(string.lower() in value.lower())

	return result


def get_current_user_info(data_type: str=None) -> Dict[str,Any]:
	'''
	Get various info about the current system user

	Properties:

	- `name: str` = The name of the user

	- `group: str` = The primary group of the user

	- `groups: list[str]` = All other groups the user is in

	- `uid: int` = ID of the user

	- `gid: int` = ID of the user's primary group

	- `gids: list[int]` = The IDs of all other groups the user is in
	'''

	data = DotDict({
		'name': getuser(),
		'group': None,
		'groups': None,
		'uid': os.getuid(),
		'gid': os.getgid(),
		'gids': os.getgroups()
	})

	try:
		data.gids.remove(data.gid)

	except:
		pass

	data.group = grp.getgrgid(data.gid).gr_name
	data.groups = [grp.getgrgid(v).gr_name for v in data.gids]

	if not data_type:
		return data

	return data[data_type]


def get_ip() -> str:
	'Get the primary IP address of the machine'

	s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

	try:
		s.connect(('10.255.255.255', 1))
		data = s.getsockname()
		ip = data[0]

	except Exception:
		ip = '127.0.0.1'

	finally:
		s.close()

	return ip


def hasher(data:Union[str,DotDict,dict,list,set,tuple,bytearray,bytes], alg='blake2s'):
	'Hash an object supported by `convert_to_bytes` and return the digest in hex format as a string. Check `hashlib.__always_supported` for valid hash algorithms'

	if alg not in hashlib.algorithms_available:
		raise TypeError('Unsupported hash algorithm. Supported algs:', ', '.join(hashlib.algorithms_available))

	if alg in ['sha1', 'md4', 'md5', 'md5-sha1']:
		logging.verbose('Warning: Using an insecure hashing algorithm. sha256 or sha512 is recommended')

	data = data.encode('UTF-8') if type(string) != bytes else data

	newhash = hashlib.new(alg)
	newhash.update(data)
	return newhash.hexdigest()


def import_from_path(mod_path:Union[str,Path]):
	'Import a module from a directory'

	mod_path = Path(mod_path)

	if mod_path.is_dir:
		path = mod_path.join('__init__.py')
		name = path.name

	else:
		path = mod_path
		name = path.name.replace('.py', '', -1)

	spec = util.spec_from_file_location(name, path)
	module = util.module_from_spec(spec)
	spec.loader.exec_module(module)
	return module


def join_list(data: Union[list,tuple,set], join_value: str = ' ') -> str:
	'Takes a list, tuple, or set, converts all items to a string, and then combines them into a single string'
	return join_value.join(str(v) for v in data)


def nfs_check(path:Union[str,Path]) -> bool:
	'''
	Check if a file or directory is on an NFS share. 

	Note: Only tested on Linux
	'''

	if platform.system() == 'Windows':
		izzylog.verbose('Refusing to check unix mounts on a non-unix system')
		return

	proc = Path('/proc/mounts')
	path = Path(path).resolve()

	if not proc.exists:
		return True

	with proc.open() as fd:
		for line in fd:
			line = line.split()

			if line[2] == 'nfs' and line[1] in path:
				return True

	return False


def port_check(port:int, address:str='127.0.0.1', tcp:bool=True) -> bool:
	'Check if a TCP or UDP port is in use'

	with socket.socket(socket.AF_INET, socket.SOCK_STREAM if tcp else socket.SOCK_DGRAM) as s:
		try:
			return s.connect_ex((address, port)) == 0

		except socket.error as e:
			return False


def print_methods(obj: object, include_underscore:bool=False):
	'Prints each method of an object on a new line. Methods starting with an underscore `_` are only printed if `include_underscore` is `True`'

	for line in dir(obj):
		if line.startswith('_'):
			if include_underscore:
				print(line)

		else:
			print(line)


def prompt(prompt:str, default:Any=None, valtype=str, options:list[Any]=[], password:bool=False) -> Any:
	'''An upgraded `input`

	Arguments: 

	- `prompt: str` = The string to display to the user

	- `default: Any` = The value that should be returned if there is no user input

	- `valtype: Callable` = The type the value should be returned as

	- `options: List[str]` = If set, these are the only values the user can select

	- `password: bool` =  If set to True, the input will be treated like a password and not show the user's input on screen
	'''

	input_func = getpass if password else input

	if default != None:
		prompt += ' [-redacted-]' if password else f' [{default}]'

	prompt += '\n'

	if options:
		options = [str(opt) for opt in options]
		opt = '/'.join(options)
		prompt += f'[{opt}]'

	prompt += ': '
	value = input_func(prompt)

	while value and len(options) > 0 and value not in options:
		input_func('Invalid value:', value)
		value = input(prompt)

	if not value or value == '':
		return default

	ret = valtype(value)

	while valtype == Path and not ret.parent.exists:
		input_func('Parent directory doesn\'t exist')
		ret = Path(input(prompt))

	return ret


def random_gen(*args, **kwargs):
	izzylog.warning('random_gen deprecated. use random_str instad.')
	return random_str(*args, **kwargs)


def random_int(lower: int = 0, upper: int = 1000000) -> int:
	return random.choice(range(lower, upper + 1))


def random_port(lower: int = 49152, upper: int = 65535, address: str = '127.0.0.1', tcp: bool = True) -> int:
	'Find a random open tcp or udp port between `upper` and `lower`.'

	selected_port = None
	cycles = 0

	while selected_port == None:
		if cycles >= 10:
			raise RuntimeError('Failed to find an open port')

		port = random.choice(range(lower, upper + 1))

		if not port_check(port, address, tcp):
			selected_port = port
			break

		cycles += 1

	return selected_port


def random_str(length:int=20, letters:bool=True, numbers:bool=True, extra:str=None) -> str:
	'Return a randomly generated string. Uses alphanumeric characters by default, but more can be specified as a string'

	characters = ''

	if letters:
		characters += string.ascii_letters

	if numbers:
		characters += string.digits

	if extra:
		characters += extra

	return ''.join(random.choices(characters, k=length))


def remove(string: str, junk: list) -> str:
	'Remove a list of strings from a string'

	for line in junk:
		string = string.replace(line, '')

	return string


def replace_strings(string, data):
	for key, value in data.items():
		string = string.replace(key, value)

	return string


def signal_handler(func:Callable, *args, original_args:bool=False, **kwargs):
	'Set a function to be called on `SIGHUP`, `SIGINT`, `SIGQUIT`, and `SIGTERM` signals. If `None` is specified for `func`, the signals are set to their default handlers.'

	if func:
		if original_args:
			handler = lambda signum, frame: func(signum, frame, *args, **kwargs)

		else:
			handler = lambda *_: func(*args, **kwargs)

	else:
		handler = signal.SIG_DFL

	signal.signal(signal.SIGHUP, handler)
	signal.signal(signal.SIGINT, handler)
	signal.signal(signal.SIGQUIT, handler)
	signal.signal(signal.SIGTERM, handler)


# This currently uses os.system instead of subprocess.Popen which I don't want
def sudo(command:str, args:str, password:str=None) -> int:
	'Run a command via sudo and return the exit code.'

	if command.startswith('/'):
		assert Path(command).isfile

	while not password:
		password = prompt(f'[PySudo] password for {getuser()}', password=True)

	#proc = Popen(['sudo', command, *shlex.split(argstring), *args], stdin=PIPE)
	#proc.communicate(bytes(password + '\n', 'utf-8'))

	#return proc

	return os.system(f'echo {password} | sudo -S -- {command} {args}')


def time_function(func:Callable, *args, passes:int=1, use_gc:bool=True, **kwargs) -> Dict[str,Union[list[float],float]]:
	'Call a function n times and return each run time, the average time, and the total time in miliseconds'

	options = [
		lambda: func(*args, **kwargs)
	]

	if use_gc:
		options.append('gc.enable()')

	timer = timeit.Timer(*options)

	if passes > 1:
		times = timer.repeat(passes, 1)

	times = [timer.timeit(1)]
	average = statistics.fmean(times)
	total = sum(times)

	return DotDict(
		runs = times,
		average = average,
		total = times
	)


def time_function_pprint(func:Callable, *args, passes:int=5, use_gc=True, floatout=True, **kwargs) -> Dict[str,Union[list[float],float]]:
	'Prints out readable results from `time_function` and returns the raw data. Convert the printed times to an `int` by setting `floatout` to `False`'

	data = time_function(func, *args, **kwargs, passes=passes, use_gc=use_gc)

	for passtime in data.times:
		if not floatout:
			print(f'Pass {idx+1}: {passtime:.0f}')

		else:
			print(f'Pass {idx+1}: {passtime:.8f}')

	print('-----------------')
	if not floatout:
		print(f'Average: {data.average:.0f}')
		print(f'Total: {data.total:.0f}')

	else:
		print(f'Average: {data.average:.8f}')
		print(f'Total: {data.total:.8f}')

	return DotDict(
		runs = times,
		average = average,
		total = sum(times)
	)


def timestamp(dtobj:datetime=None, utc:bool=False) -> float:
	'Convert a `datetime` object to a unix timestamp.'

	dtime = dtobj if dtobj else datetime
	date = dtime.utcnow() if utc else dtime.now()

	return date.timestamp()


def var_name(**kwargs):
	'Return multiple variable names as a string'

	return list(kwargs.keys())


class ArgParser(argparse.ArgumentParser):
	'[WIP] Wrapper around argparse.ArgumentParser to make parsing of sys.argv easier'
	def __init__(self):
		super().__init__()

		self._args = None
		self._arguments = DotDict()


	def __getitem__(self, key):
		return self._arguments[key]


	def __setitem__(self, key, value):
		if not isinstance(value, dict):
			raise TypeError(f'Value must be a dict, not {class_name(value)}')

		self._arguments[key] = Argument(**value)


	def add_positional(self, name, **kwargs):
		self[name] = Argument(name=name, **kwargs)


	def add_optional(self, name, short_name=None, **kwargs):
		self[name] = Argument(name=f'--{name}', short_name=f'-{short_name}', **kwargs)


	def add_action(self, name, callback, *args, argument=None, **kwargs):
		if argument:
			self[name] = argument

		self[name].set_callback(callback, *args, **kwargs)


	@property
	def args(self):
		if not self._args:
			self._args = self.parse_args()

		return self._args


class Argument(DotDict):
	'An `ArgParser` argument'

	valid_keys = ['name', 'short_name', 'nargs', 'default', 'type', 'help', 'metavar']
	callback = None

	def __setitem__(self, key, value):
		if key not in self.valid_keys:
			raise KeyError(f'Not a valid argument option: {key}')

		if not value:
			return

		if key == 'name':
			if not self.metavar:
				super().__setitem__('metavar', value)

			value = value.replace(' ', '-', '_', '-')

		super().__setitem__(key, value)


	def set_callback(self, callback, *args, **kwargs):
		self.callback = lambda *cliargs: callback(*cliargs, *args, **kwargs)


class Async:
	def __init__(self, function, *args):
		if not asyncio.iscoroutinefunction(function):
			raise TypeError(f'Function must be an awaitable')

		self.func = function
		self.obj = None
		self.cls = None


	def __call__(self, *args, **kwargs):
		if self.obj:
			return self.func(self.obj, *args, **kwargs)

		return self.func(*args, **kwargs)


	def __get__(self, obj, cls):
		self.obj = obj
		self.cls = cls

		return self


	def sync(self, *args, **kwargs):
		result = asyncio.run(self(*args, **kwargs))
		return result


class Boolean(int):
	def __new__(cls, value):
		if not isinstance(value, (str, bool, int, type(None))):
			raise TypeError(f'Value is not a string, boolean, int, or nonetype: {value}')

		if isinstance(value, str):
			if value.lower() in ['on', 'y', 'yes', 'true', 'enable', 'enabled']:
				value = True

			elif value.lower() in ['off', 'n', 'no', 'false', 'disable', 'disable']:
				value = False

			else:
				raise TypeError(f'Cannot parse string "{value}" as a boolean')

		else:
			try:
				value = value.__bool__()

			except AttributeError:
				raise TypeError(f'Cannot convert object of type "{class_name(value)}"')

		return int.__new__(cls, value)


	def __repr__(self):
		return f'Boolean({self.raw()})'


	def __str__(self):
		return str(self.raw())


	def raw(self):
		return bool(self)


# compat
boolean = convert_to_boolean
