from .dotdict import DotDict
from .exceptions import ReadOnlyError
from .misc import class_name


class ObjectBase:
	__props = {}
	__defaults = {}

	def __init__(self, readonly_props=[], **kwargs):
		self.__props = dict()
		self.__defaults = kwargs
		self.__readonly = []

		for key, value in kwargs.items():
			self.set_property(key, value)

		self.set_readonly(readonly_props)


	def __repr__(self):
		attrs = ', '.join(f'{key}={repr(value)}' for key, value in self.__props.items())
		return f'{class_name(self)}({attrs})'


	def __getattr__(self, key):
		if key not in self.__props:
			return object.__getattribute__(self, key)

		return self.__props[key]


	def __setattr__(self, key, value):
		if key not in self.__props:
			return object.__setattr__(self, key, value)

		if self.__readonly == True or key in self.__readonly:
			raise ReadOnlyError(key)

		self.__props[key] = self._parse_property(key, value)


	def __delattr__(self, key):
		if key not in self.__props:
			return object.__delattr__(self, key, value)

		if self.__readonly == True or key in self.__readonly:
			raise ReadOnlyError(key)

		del self.__props[key]


	def __getitem__(self, key):
		return self.__props[key]


	def __setitem__(self, key, value):
		self.__props[key] = value


	def __delitem__(self, key):
		del self.__props[key]


	def __contains__(self, key):
		return key in self.keys()


	def __iter__(self):
		return self.__props.__iter__()


	def _parse_property(self, key, value):
		return value


	def items(self):
		return self.__props.items()


	def keys(self):
		return tuple(self.__props.keys())


	def values(self):
		return tuple(self.__props.values())


	def get_default(self, key):
		return self.__defaults[key]


	def get_property(self, key):
		return self.__props[key]


	def set_default(self, key):
		self.set_property(key, self.__defaults[key])


	def set_property(self, key, value):
		self.__props[key] = self._parse_property(key, value)


	def unset_property(self, key):
		del self.__props[key]


	def set_readonly(self, *props):
		self.__readonly = True if props == True else tuple(props)


	def to_dict(self, *exclude):
		return DotDict({k:v for k,v in self.items() if k not in exclude})


	def to_json(self, *args, **kwargs):
		return self.to_dict().to_json(*args, **kwargs)

