import json
import mimetypes
import os
import platform
import re
import shutil
import sys

from datetime import datetime
from functools import cached_property
from glob import iglob

from . import izzylog
from .enums import FileSizeUnit, FileSizeMulti, PathType

try:
	import grp
	import pwd

except ImportError:
	grp = None
	pwd = None

try:
	from magic import Magic
	magic = Magic(uncompress=True, mime=True)

except ImportError:
	magic = None


linux_prefix = dict(
	bin = '.local/bin',
	share = '.local/share',
	cache = '.cache',
	config = '.config'
)


class TinyDotDict(dict):
	__getattr__ = dict.__getitem__
	__setattr__ = dict.__setitem__
	__delattr__ = dict.__delitem__


class FileSize(int):
	def __new__(cls, size, unit='B'):
		if isinstance(size, str):
			size = int(size)

		elif not isinstance(size, (int, float)):
			raise TypeError(f'Size must be an int, float, or string representation of an int')

		unit = FileSizeUnit.parse(unit)
		multi = FileSizeMulti.parse(unit.name)

		return int.__new__(cls, size * multi)


	def __repr__(self):
		value = int(self)
		return f'FileSize({value:,} bytes)'


	def __str__(self):
		return str(int(self))


	@classmethod
	def new_from_string(cls, string):
		size, unit = string.strip().split(' ', 1)
		size = float(size)
		unit = FileSizeUnit.parse(unit)
		multi = FileSizeMulti.parse(unit.name)

		return cls(int(size * multi))


	def to_fuzzy(self):
		if self < FileSize(1, 'K'):
			return 'teeny tiny'

		elif self < FileSize(1, 'M'):
			return 'smol uwu bean'

		elif self < FileSize(512, 'M'):
			return 'average'

		elif self < FileSize(1, 'G'):
			return 'big boy'

		elif self < FileSize(1, 'T'):
			return 'heckin chonker'

		elif self >= FileSize(1, 'T'):
			return 'mega chonk'


	def to_optimal_string(self):
		index = 0
		size = int(self)

		while True:
			if size < 1024 or index == 8:
				unit = FileSizeUnit.from_index(index)
				return f'{size:.2f} {unit}'

			try:
				index += 1
				size = self / FileSizeMulti.from_index(index)

			except IndexError:
				raise ValueError('File size is too large to convert to a string') from None


	def to_string(self, multi, decimals=2):
		multi = FileSizeMulti.parse(multi)

		if multi == FileSizeMulti.BYTE:
			return f'{self} B'

		size = round(self / multi, decimals)
		unit = FileSizeUnit.parse(multi.name)

		return f'{size} {unit}'


class FileStats:
	def __init__(self, path):
		self.__path = path
		self.__stats = None

		self.refresh()


	def __repr__(self):
		return f'FileStats("{self.__path}", size="{self.size:,}", mode={str(self.mode)[-3:]})'


	@property
	def atime(self):
		return datetime.fromtimestamp(self.__stats.st_atime)


	@property
	def blocks(self):
		return self.__stats.st_blocks


	@property
	def ctime(self):
		return datetime.fromtimestamp(self.__stats.st_ctime)


	@property
	def devid(self):
		return self.__stats.st_dev


	@property
	def gid(self):
		return self.__stats.st_gid


	@property
	def group(self):
		if not grp:
			raise NotImplementedError('grp module not available')

		return grp.getgrgid(self.gid).gr_name


	@property
	def inode(self):
		return self.__stats.st_ino


	@property
	def logical_size(self):
		return FileSize(self.__stats.st_size)


	@property
	def mtime(self):
		return datetime.fromtimestamp(self.__stats.st_mtime)


	@property
	def mode(self):
		return UnixPerms.new_from_st_mode(self.__stats.st_mode)


	@property
	def real_size(self):
		return FileSize(self.blocks * 512)


	@property
	def size(self):
		try:
			return self.real_size

		except AttributeError:
			return self.logical_size


	@property
	def uid(self):
		return self.__stats.st_uid


	@property
	def user(self):
		if not pwd:
			raise NotImplementedError('pwd module not available')

		return pwd.getpwuid(self.uid).pw_name


	def refresh(self):
		self.__stats = os.stat(self.__path)


class UnixPerms:
	PERMS = TinyDotDict(
		READ = 4,
		WRITE = 2,
		EXECUTE = 1
	)

	SHORT_PERMS = TinyDotDict(
		R = 4,
		W = 2,
		X = 1
	)

	READ = PERMS.READ
	WRITE = PERMS.WRITE
	EXECUTE = PERMS.EXECUTE

	R = SHORT_PERMS.R
	W = SHORT_PERMS.W
	X = SHORT_PERMS.X

	KEYS = ['user', 'group', 'other']


	def __init__(self, user=6, group=4, other=4):
		self.__perms = [0,0,0]

		self['user'] = user
		self['group'] = group
		self['other'] = other


	def __repr__(self):
		return f'{self.__class__.__name__}(user={self[0]}, group={self[1]}, other={self[2]})'


	def __str__(self):
		values = ''.join(str(perm) for perm in self.__perms)
		return f'0o{values}'


	def __int__(self):
		return int(self.__str__()[2:], 8)


	def __iter__(self):
		for perm in self.__perms:
			yield perm


	def __getitem__(self, key):
		if isinstance(key, str):
			key = self.KEYS.index(key)

		elif not isinstance(key, int):
			raise TypeError('Key must be a str or int')

		assert key < 3
		return self.__perms[key]


	def __setitem__(self, key, value):
		if isinstance(key, str):
			key = self.KEYS.index(key)

		if isinstance(value, str):
			value = self._str_to_int(value)

		self.__perms[key] = value


	@classmethod
	def parse(cls, data):
		if isinstance(data, UnixPerms):
			return cls(data['user'], data['group'], data['other'])

		if isinstance(data, int):
			try:
				data = int(str(data), 8)

			except ValueError:
				pass

			data = format(data, 'o')
			return cls(*(int(n) for n in data))

		elif isinstance(data, str):
			return cls(data[0:3], data[3:6], data[6:9])

		raise TypeError('Data must be an int, oct, or str')


	@classmethod
	def new_from_st_mode(cls, data):
		data = oct(data)[-3:]

		return cls(*[int(v) for v in data])


	def _int_to_str(self, value):
		new_value = ''

		for name, num in self.SHORT_PERMS.items():
			if num <= value:
				new_value += name.lower()
				value -= num

			else:
				new_value += '-'

		return new_value


	def _str_to_int(self, string):
		assert isinstance(string, str)

		new_value = 0

		for value in string:
			if value == '-':
				continue

			new_value += self.SHORT_PERMS[value.upper()]

		return new_value


	def to_int(self):
		return self.__int__()


	def to_string(self):
		return ''.join(self._int_to_str(perm) for perm in self)


class PathMeta(type):
	@property
	def cache(cls):
		return cls.new_cache_path()


	@property
	def cwd(cls):
		return cls.new_cwd()


	@property
	def home(cls):
		return cls.new_home()


	@property
	def script(cls):
		return cls.new_script_path()


	@property
	def xdg(cls):
		home = cls.home

		return TinyDotDict(
			cache = os.environ.get('XDG_CACHE_HOME', home.join('.cache')),
			config = os.environ.get('XDG_CONFIG_HOME', home.join('.cache')),
			data = os.environ.get('XDG_DATA_HOME', home.join('.local/share')),
			runtime = os.environ.get('XDG_RUNTIME_DIR', f'/run/user/{os.getuid()}'),
			state = os.environ.get('XDG_DATA_STATE', home.join('.local/state'))
		)


	def module(cls, module):
		return cls(module.__file__).parent


	def app_data_dir(cls, prefix:str, author:str, name:str):
		'Returns the cache or config path for software'

		assert prefix in ['config', 'cache']

		if platform.system() == 'Linux':
			paths = dict(
				config = cls('~/.config'),
				cache = cls('~/.cache')
			)

		elif platform.system() == 'Darwin':
			paths = dict(
				config = cls('~/Application Support'),
				cache = cls('~/Library/Caches')
			)

		elif platform.system() == 'Windows':
			return dict(
				config = cls('~/Application Data/Local Settings').join(author, name),
				cache = cls('~/Application Data').join(author, name, 'Cache')
			)[prefix]

		else:
			raise TypeError(f'Unknown system: {platform.system()}')

		return paths[prefix].join(author, name)


	def which(cls, name):
		return cls(shutil.which(name))


	# Made this before I realized shutil.which existed. Keeping just in case.
	def which2(cls, name):
		for path in os.environ.get('PATH', '').split(':'):
			match = Path(path).join(name)

			if match.exists():
				return match

		raise FileNotFoundError(f'Cannot find "{name}" in PATH')


	def unix(cls, prefix, author, software):
		try:
			return cls.home.join(linux_prefix[prefix], author, software)

		except KeyError:
			raise KeyError(f'Not a valid Linux user directory prefix: {prefix}')


	def windows(cls, prefix, author, software):
		pass


class Path(str, metaclass=PathMeta):
	def __init__(self, path=os.getcwd(), exist=True, missing=True, parents=True):
		## todo: move these to direct properties of Path
		self.config = TinyDotDict({
			'missing': missing,
			'parents': parents,
			'exist': exist
		})

		self.__stat = None


	def __new__(cls, path, resolve=False):
		if str(path).startswith('~'):
			new_path = str.__new__(cls, os.path.expanduser(path))

		else:
			new_path = str.__new__(cls, path)

		if resolve:
			return new_path.resolve()

		return new_path


	def __repr__(self):
		return f'Path("{self}")'


	def __enter__(self):
		self.fd = self.open('r')
		return self.fd


	def __exit__(self, *args):
		self.fd.close()


	def __check_dir(self, path=None):
		target = self if not path else Path(path)

		if not self.config.parents and not target.parent.exists():
			raise FileNotFoundError('Parent directories do not exist:', target)

		if not self.config.exist and target.exists():
			raise FileExistsError('File or directory already exists:', target)


	@classmethod
	def new_cache_path(cls):
		try:
			return cls(os.environ['XDG_CACHE_HOME'])

		except KeyError:
			return cls.home.join('.cache')


	@classmethod
	def new_cwd(cls, path=None):
		new_path = cls(os.getcwd()).resolve()

		if path:
			return new_path.join(path)

		return new_path


	@classmethod
	def new_home(cls, path=None):
		new_path = cls('~')

		if path:
			return new_path.join(path)

		return new_path


	@classmethod
	def new_script_path(cls):
		try: path = sys.modules['__main__'].__file__
		except: path = sys.argv[0]

		return cls(path).parent


	@cached_property
	def name(self):
		return os.path.basename(self)


	@cached_property
	def parent(self):
		return Path(os.path.dirname(self))


	@cached_property
	def stem(self):
		return os.path.splitext(self.name)[0]


	@cached_property
	def suffix(self):
		return os.path.splitext(self.name)[1][1:]


	def accessed(self):
		return self.stat().atime


	def append(self, text):
		return Path(self + text)


	def backup(self, ext='backup', overwrite=False):
		target = f'{self}.{ext}'
		self.copy(target, overwrite)
		return Path(target)


	def chmod(self, mode):
		if not mode:
			mode = UnixPerms(6, 4, 4)

		elif not isinstance(mode, UnixPerms):
			mode = UnixPerms.parse(mode)

		os.chmod(self, mode.to_int())


	def contains(self, text):
		return text in str(self)


	def copy(self, path, overwrite=False):
		target = Path(path)

		self.__check_dir(path)

		if target.exists():
			if overwrite:
				try:
					target.delete()

				except FileNotFoundError:
					pass

			elif not self.config.exist:
				raise FileExistsError(target)

		shutil.copy2(self, target)
		return target


	def create(self, file=True, mode=None):
		if self.exists():
			if self.config.exist:
				return

			raise FileExistsError(self)

		if not mode:
			mode = UnixPerms(6,4,4) if file else UnixPerms(7,5,5)

		os.makedirs(self.parent, mode.to_int(), exist_ok=self.config.exist)

		if file:
			with self.open('w+') as fd:
				pass

		else:
			os.makedirs(self, mode.to_int(), exist_ok=self.config.exist)

		return self.exists()


	def created(self):
		return self.stat().ctime


	def delete(self):
		if not self.exists():
			if not self.config.exist:
				raise FileNotFoundError(self)

			return

		if self.isdir():
			shutil.rmtree(self, ignore_errors=True)

		else:
			os.remove(self)

		return not self.exists


	def exists(self):
		return os.path.exists(self)


	def expanduser(self):
		izzylog.warning(f'Path.expanduser does not actually do anything and will be removed soon.')
		return self


	def fallocate(self, size, multi=FileSizeMulti.MEBIBYTE, mode=None):
		if not isinstance(size, int):
			raise TypeError(f'Size must be an int, not {type(size).__name__}')

		multi = FileSizeMulti.parse(multi)

		if not self.exists():
			self.touch(mode=mode)

		fullsize = size * multi

		with self.open('ab') as fd:
			fd.truncate(fullsize)


	def get_type(self):
		types = []

		if os.path.isfile(self):
			types.append(PathType.FILE)

		elif os.path.isdir(self):
			types.append(PathType.DIR)

		else:
			types.append(PathType.UNKNOWN)

		if os.path.islink(self):
			types.append(PathType.LINK)

		return tuple(types)


	def glob(self, pattern='*', recursive=True):
		if recursive and not pattern.startswith('**/'):
			pattern = '**/' + pattern

		for path in iglob(pattern, root_dir=self, recursive=recursive):
			yield self.join(path)


	def group(self):
		return self.stat().group


	def groupid(self):
		return self.stat().gid


	def isdir(self):
		return self.istype(PathType.DIR)


	def isfile(self):
		return self.istype(PathType.FILE)


	def islink(self):
		return self.istype(PathType.LINK)


	def istype(self, file_type):
		return file_type in self.get_type()


	def join(self, *paths, case_sensitive=True):
		if not len(paths):
			raise ValueError('Missing path to join')

		if case_sensitive:
			return Path(os.path.join(self, *paths))

		old_path = self
		new_path = None

		if len(paths) == 1:
			paths = os.path.split(paths[0])

		for path in paths:
			print(old_path, path)
			try:
				new_path = list(old_path.search(path, case_sensitive=False, recursive=False))[0]
				old_path = new_path

			except IndexError:
				raise FileNotFoundError(f'Cannot find path at {os.path.join(self, *paths)}') from None

		return new_path


	def json_load(self):
		with self as s:
			return json.load(s)


	def json_dump(self, data, indent=None):
		with self.open('w') as s:
			s.write(json.dumps(data, indent=indent))


	def link(self, path):
		target = Path(path)

		self.__check_dir(path)

		if target.exists():
			if not self.config.exist:
				raise FileExistsError(target)

			target.delete()

		self.symlink_to(path, target.isdir())
		return target


	def listdir(self, recursive=False, symlinks=False):
		for path in os.listdir(self):
			path = self.join(path)

			if not symlinks and path.islink():
				continue

			yield path

			if recursive and path.isdir():
				for path in path.listdir(recursive, symlinks):
					yield path


	def mime(self):
		if magic:
			return magic.from_file(self)

		return mimetypes.guess_type(self)[0]


	def mkdir(self, mode=None):
		return self.create(file=False, mode=mode)


	def mkparent(self, mode=None):
		self.parent.mkdir(mode)


	def modified(self):
		return self.stat().mtime


	def move(self, path, overwrite=False):
		dest = Path(path)

		if not overwrite and dest.exists:
			raise FileExistsError(f'Refusing to move file to existing destination: {dest}')

		shutil.move(self, dest)


	def open(self, *args, **kwargs):
		return open(self, *args, **kwargs)


	def permissions(self):
		return self.stat().mode


	def read(self, byte=False):
		with self.open('rb' if byte else 'r') as fd:
			return fd.read()


	def readlines(self):
		with self.open() as fd:
			for line in fd.readlines():
				yield line


	def resolve(self):
		path = Path(os.path.abspath(self))

		try:
			if path.islink():
				return Path(os.path.realpath(path))

		except:
			pass

		return path


	def search(self, text, case_sensitive=True, recursive=False, symlinks=False):
		for path in self.listdir(recursive, symlinks):
			if not case_sensitive and text.lower() in path.name.lower():
				yield path

			elif case_sensitive and text in path.name:
				yield path


	def size(self, recursive=False, symlinks=False):
		if recursive and self.isdir():
			size = 0

			for path in self.listdir(recursive, symlinks):
				size += path.stat().size

			return FileSize(size)

		return self.stat().size


	def stat(self, refresh=False):
		if not self.__stat:
			self.__stat = FileStats(self)

		elif refresh:
			self.__stat.refresh()

		return self.__stat


	def touch(self, mode=None, mtime=None):
		if mtime and not isinstance(mtime, datetime):
			raise TypeError(f'mtime must be a datetime object or None')

		self.create(file=True)

		timestamp = mtime or datetime.now().timestamp()
		os.utime(self, (timestamp, timestamp))


	def user(self):
		return self.stat().user


	def userid(self):
		return self.stat().uid


	def write(self, data, mode='w'):
		with self.open(mode) as fd:
			fd.write(data)
