import os
import shlex
import subprocess
import time

from . import izzylog
from .dotdict import DotDict
from .misc import Boolean, class_name
from .path import Path


class ProcessMeta(type):
	def ldd(cls, binary):
		if not os.path.split(binary)[0]:
			binary = Path.which(binary)

		else:
			binary = Path(binary).resolve()

		proc = cls(Path.which('ldd'), binary)
		proc.set_pipe('stdout')
		proc.run()

		if proc.returncode != 0:
			raise ValueError(f'ldd exit with error code: {proc.returncode}')

		libs = DotDict()

		for line in proc.readlines('stdout'):
			line = line.split()

			if len(line) == 2:
				if not line[0].startswith('/'):
					continue

				key = Path(line[0]).name
				value = line[0]

			elif len(line) == 4:
				if '=>' not in line:
					continue

				key = line[0]
				value = line[2]

			else:
				izzylog.debug('ignore:', line)
				continue

			libs[key.replace('.so.', '')] = value

		return libs


	def new_python_command(cls, *args, **kwargs):
		return cls(sys.executable, *args, **kwargs)


	def new_python_module(cls, module, *args, **kwargs):
		return cls(sys.executable, '-m', module, *args, **kwargs)


class Process(metaclass=ProcessMeta):
	defaults = {
		'executable': None,
		'args': None,
		'stdin': None,
		'stdout': None,
		'stderr': None,
		'include_env': True,
		'env': {},
		'shell_executable': Path.which('bash'),
		'shell': False,
		'cwd': Path.new_cwd(),
		'user': None,
		'group': None,
		'extra_groups': None
	}

	def __init__(self, executable, *args, run=False, **kwargs):
		self.__props = {}
		self.__previous_exit = None
		self.proc = None

		self.executable = Path(executable)
		self.args = ' '.join(args)

		for key, value in kwargs.items():
			self[key] = value

		if run:
			self.run()


	def __getattr__(self, key):
		if key not in self.defaults:
			return object.__getattribute__(self, key)

		return self.__getitem__(key)


	def __setattr__(self, key, value):
		if key not in self.defaults:
			return object.__setattr__(self, key, value)

		self.__setitem__(key, value)


	def __delattr__(self, key):
		if key not in self.defaults:
			return object.__delattr__(self, key, value)

		self.__delitem__(key)


	def __getitem__(self, key):
		if key not in self.defaults:
			raise KeyError(key)

		try:
			return self.__props[key]

		except KeyError:
			return self.defaults[key]


	def __setitem__(self, key, value):
		if key not in self.defaults:
			raise KeyError(key)

		if not value:
			self.__delitem__(key)

		self.__props[key] = self._parse_property(key, value)


	def __delitem__(self, key):
		if key not in self.defaults:
			raise KeyError(key)

		try:
			del self.__props[key]

		except KeyError:
			return


	def __enter__(self):
		self.start()

		return self


	def __exit__(self, *args):
		self.stop()


	def __del__(self):
		self.stop()


	def _parse_property(self, key, value):
		if key in ['executable', 'shell_executable', 'cwd']:
			path = Path(value)

			if not path.exists():
				return Path.which(value)

			return path

		elif key == 'args' and not isinstance(value, str):
			if isinstance(value, (list, set, tuple)):
				return ' '.join(value)

			raise TypeError(f'args must be a str, list, set, or tuple, not {class_name(value)}')

		elif key in ['include_env', 'shell']:
			return Boolean(value)

		elif key == 'env':
			return DotDict(value)

		return value


	@property
	def returncode(self):
		if not self.proc:
			return False

		return self.proc.poll()


	@property
	def running(self):
		if self.returncode == None:
			return True

		return False


	def set_pipe(self, key):
		if key not in ['stdin', 'stdout', 'stderr']:
			raise KeyError('Pipe name must be "stdin", "stdout", "stderr"')

		self[key] = subprocess.PIPE


	def unset_pipe(self, key):
		if key not in ['stdin', 'stdout', 'stderr']:
			raise KeyError('Pipe name must be "stdin", "stdout", "stderr"')

		del self[key]


	def read(self, key, length=None, decode=True):
		if key not in ['stdout', 'stderr']:
			raise KeyError(f'Key must be "stdout" or "stderr"')

		if self[key] != subprocess.PIPE:
			raise FileNotFoundError(f'Cannot access {key} since it was not set to subprocess.PIPE')

		output = getattr(self.proc, key)
		data = output.read(length) if length else output.read()

		return data.decode('utf-8') if decode else data


	def readlines(self, key, decode=True):
		if key not in ['stdout', 'stderr']:
			raise KeyError(f'Key must be "stdout" or "stderr"')

		if self[key] != subprocess.PIPE:
			raise FileNotFoundError(f'Cannot access {key} since it was not set to subprocess.PIPE')

		output = getattr(self.proc, key)

		for line in output.readlines():
			yield line.decode('utf-8') if decode else line


	def write(self, *message):
		if self.stdin != subprocess.PIPE:
			raise FileNotFoundError(f'Cannot access stdin since it was not set to subprocess.PIPE')

		msg = ' '.join(message)
		self.proc.stdin.write(msg.encode('utf-8'))


	def start(self):
		if self.running:
			return

		command = [self.executable, *shlex.split(self.args)]

		kwargs = DotDict(
			stdin = self.stdin,
			stdout = self.stdout,
			stderr = self.stderr,
			env = {},
			cwd = self.cwd,
			user = self.user,
			group = self.group,
			extra_groups = self.extra_groups
		)

		if self.shell:
			kwargs.shell = True
			kwargs.executable = self.shell_executable

		if self.include_env:
			kwargs.env.update(os.environ)

		kwargs.env.update(self.env)

		self.proc = subprocess.Popen(command, **kwargs)


	def stop(self, timeout=5):
		if not self.running:
			return self.returncode

		self.proc.terminate()

		try:
			self.proc.wait(timeout)

		except subprocess.TimeoutExpired:
			self.proc.kill()

			try:
				self.proc.wait(timeout)

			except subprocess.TimeoutExpired:
				pass

		return self.returncode


	def restart(self, timeout=5):
		self.stop(timeout)
		self.start()


	def wait(self, delay=1):
		assert self.running, 'Process needs to be running'

		while self.proc.poll() == None:
			time.sleep(delay)

		return self.stop()


	def run(self, delay=1):
		self.start()
		return self.wait(delay)


	def to_dict(self):
		return DotDict(**self.__props)
