import platform, vdf

from datetime import datetime

from . import izzylog
from .dotdict import DotDict
from .path import Path


data_paths = DotDict(
	Linux = [
		'~/.steam',
		'~/.local/share/Steam',
		'~/.local/share/steam'
	],
	Windows = [
		'C:/Program Files (x86)/Steam',
		'C:/Program Files/Steam'
	],
	Darwin = [
		'~/Library/Application Support/Steam'
	]
)

size_formats = {
	'K': 1024,
	'M': 1024**2,
	'G': 1024**3
}


def get_data_path():
	try: library_paths = data_paths[platform.system()]
	except KeyError: raise OSError(f'Unsupported operating system: {platform.system()}')

	for path in library_paths:
		for cfg_path in Path(path).resolve().glob('config/config.vdf'):
			return cfg_path.parent.parent

	raise FileNotFoundError('Cannot find Steam data path')


def get_libraries(datapath=None):
	path = Path(datapath or get_data_path()).join('config', 'libraryfolders.vdf')

	if not path.exists():
		raise FileNotFoundError(f'Cannot find libraryfolders.vdf: {path}')

	steam_dirs = []

	with path.open() as fd:
		for _, data in vdf.load(fd)['libraryfolders'].items():
			try: steam_dirs.append(Library(data['path']))
			except TypeError: pass

	return steam_dirs


class Library(Path):
	def __new__(cls, path):
		for subpath in ['steamapps', 'SteamApps']:
			try:
				return Path.__new__(cls, Path(path).join(subpath, exist=False))

			except FileNotFoundError:
				pass

		print('heck')
		raise FileNotFoundError(f'Cannot find steam library')


	def __getitem__(self, key):
		game = self.game(key)

		if not game:
			raise KeyError(f'Cannot find game with ID: {key}')

		return game


	def __iter__(self):
		return self.games()


	def games(self):
		for manifest_file in self.glob('*.acf', False):
			yield Game(self, manifest_file)


	def games_sorted(self):
		return sorted(self.games(), key=lambda x:x.name)


	def game(self, gid):
		for game in self.games():
			if game.id == gid:
				return game


class Game(DotDict):
	def __init__(self, library, manifest_path):
		with manifest_path.open() as fd:
			manifest = vdf.load(fd)['AppState']

		super().__init__({
			'id': int(manifest['appid']),
			'name': manifest['name'],
			'directory': library.join('common', manifest['installdir']),
			'updated': datetime.fromtimestamp(int(manifest['LastUpdated'])),
			'install_size': int(manifest['SizeOnDisk']),
			'compat': library.join('compatdata', manifest['appid'])
		})


	def __str__(self):
		return f'{self.name} ({self.id})'


	@property
	def path(self):
		return self.directory


	@property
	def prefix(self):
		if self.compat.exists():
			return ProtonPrefix(self.compat)


	def size(self, format=None):
		if format:
			calculated_size = self.install_size / size_formats[format]

		else:
			calculated_size = self.install_size / size_formats['G']
			format = 'G'

			if calculated_size < 1:
				calculated_size = self.install_size / size_formats['M']
				format = 'M'

			if calculated_size < 1:
				calculated_size = self.install_size / size_formats['K']
				format = 'K'

		return f'{round(calculated_size, 2)} {format}iB'


class ProtonPrefix(Path):
	def __new__(cls, path):
		if not Path(path).pfx.exists():
			raise FileNotFoundError(f'Not a proton prefix: {path}')

		return Path.__new__(cls, path)


	@property
	def version(self):
		try:
			with self.join('version') as fd:
				return fd.read().strip()

		except FileNotFoundError:
			return False


class CompatibilityTools(Path):
	def __new__(cls, path):
		if not Path(path).endswith('compatibilitytools.d'):
			raise FileNotFoundError(f'Not a compatibility tools directory: {path}')

		return Path.__new__(cls, path)


	def versions(self):
		for path in self.listdir(recursive=False):
			with path.version as fd:
				version


class Proton(Path):
	def __init__(self, path):
		with self.join('version') as fd:
			self.version = fd.read().strip()


	def __new__(cls, path):
		if not Path(path).version.exists():
			raise FileNotFoundError(f'Not a Proton directory: {path}')

		return Path.__new__(cls, path)
