import ipaddress
import mimetypes
import re
import socket

from datetime import datetime, timedelta
from functools import cached_property
from urllib.parse import urlparse, quote
from urllib.request import urlopen

from . import izzylog
from .dotdict import DotDict
from .enums import IpAddressType, IpAddressVersion, Protocol
from .exceptions import DNSResolverError, InvalidAddressError
from .path import Path

# typing modules
from typing import Union


__all__ = ['Url', 'Address']
__pdoc__ = {
	'Url.anchor': 'Text after the "#" at the end of the url',
	'Url.host': 'The domain and, if set, port as a string',
	'Url.password': 'Password portion of the url',
	'Url.path': 'Path portion of a url',
	'Url.port': 'Port number for the url. It one is not specified, it will be guessed based on the protocol',
	'Url.proto': 'Protocol portion of the url',
	'Url.query': 'Query key/value pairs as a dict',
	'Url.query_string': 'Query key/value pairs as a string',
	'Url.username': 'Username portion of the url'
}

IPV4 = re.compile(r'^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$')
IPV6 = re.compile(r'^(?:(?:[0-9A-Fa-f]{1,4}:){6}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|::(?:[0-9A-Fa-f]{1,4}:){5}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:[0-9A-Fa-f]{1,4})?::(?:[0-9A-Fa-f]{1,4}:){4}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4})?::(?:[0-9A-Fa-f]{1,4}:){3}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:(?:[0-9A-Fa-f]{1,4}:){,2}[0-9A-Fa-f]{1,4})?::(?:[0-9A-Fa-f]{1,4}:){2}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:(?:[0-9A-Fa-f]{1,4}:){,3}[0-9A-Fa-f]{1,4})?::[0-9A-Fa-f]{1,4}:(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:(?:[0-9A-Fa-f]{1,4}:){,4}[0-9A-Fa-f]{1,4})?::(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:(?:[0-9A-Fa-f]{1,4}:){,5}[0-9A-Fa-f]{1,4})?::[0-9A-Fa-f]{1,4}|(?:(?:[0-9A-Fa-f]{1,4}:){,6}[0-9A-Fa-f]{1,4})?::)$')
DOMAIN = re.compile(r'^([A-Za-z0-9]\.|[A-Za-z0-9][A-Za-z0-9-]{0,61}[A-Za-z0-9]\.){1,3}[A-Za-z]{2,6}$')


class Address(str):
	def __new__(cls, address, timeout=5):
		hostname = None

		if not (IPV4.match(address) or IPV6.match(address)):
			hostname = address
			default_timeout = socket.getdefaulttimeout()

			try:
				socket.setdefaulttimeout(timeout)
				address = socket.gethostbyname(address)

			except socket.gaierror as e:
				if e.errno in [-2, -3, -5]:
					raise DNSResolverError(address) from None

				raise e from None

			finally:
				socket.setdefaulttimeout(default_timeout)

		try:
			ip_addr = ipaddress.ip_address(address)

		except ValueError:
			InvalidAddressError(address)

		addr = str.__new__(cls, address)
		addr._addr = ip_addr
		addr._hostname = hostname

		return addr


	def __repr__(self):
		if self.hostname:
			return f'Address({self}, type={self.type}, version={self.version.value}, hostname={self.hostname})'

		return f'Address({self}, type={self.type}, version={self.version.value})'


	@cached_property
	def type(self) -> IpAddressType:
		if self._addr.is_loopback:
			return IpAddressType.LOOPBACK

		elif self._addr.is_reserved:
			return IpAddressType.RESERVED

		elif self._addr.is_link_local:
			return IpAddressType.LOCAL

		elif self._addr.is_multicast:
			return IpAddressType.MULTICAST

		elif self._addr.is_global:
			return IpAddressType.PUBLIC

		elif self._addr.is_private:
			return IpAddressType.PRIVATE

		return IpAddressType.UNKNOWN


	@cached_property
	def loopback(self):
		return self.is_type(IpAddressType.LOOPBACK)


	@cached_property
	def reserved(self):
		return self.is_type(IpAddressType.RESERVED)


	@cached_property
	def local(self):
		return self.is_type(IpAddressType.LOCAL)


	@cached_property
	def multicast(self):
		return self.is_type(IpAddressType.MULTICAST)


	@cached_property
	def public(self):
		return self.is_type(IpAddressType.PUBLIC)


	@cached_property
	def private(self):
		return self.is_type(IpAddressType.PRIVATE)


	@cached_property
	def unknown(self):
		return self.is_type(IpAddressType.UNKNOWN)


	@cached_property
	def version(self):
		if isinstance(self._addr, ipaddress.IPv4Address):
			return IpAddressVersion.IPV4

		elif isinstance(self._addr, ipaddress.IPv6Address):
			return IpAddressVersion.IPV6

		raise TypeError(f'Unknown IP address version: {class_name(self._addr)}')


	@property
	def hostname(self):
		return self._hostname


	def is_type(self, *iptypes):
		for iptype in iptypes:
			if not isinstance(iptype, IpAddressType):
				try:
					iptype = IpAddressType[iptype.upper()]

				except KeyError:
					raise KeyError(f'Invalid address type: {iptype}')

			if self.type == iptype:
				return True

		return False


class Url(str):
	'`str` representation of a url parsed with urlparse'

	proto:str = None
	host:str = None
	port:int = None
	path:Path = None
	query_string:str = None
	query:DotDict = None
	username:str = None
	password:str = None
	anchor:str = None


	def __init__(self, url):
		self._tldcache = None
		self._tldcache_path = Path.cache.join('icann_public_suffix_list.txt')
		self._address = None

		if not isinstance(url, str):
			raise TypeError('url is not a string')

		self._parsed = urlparse(url)

		#if not all([self.proto, self.domain]):
			#raise TypeError('Not a valid url')


	@classmethod
	def new(cls, domain:str, path:Union[Path,str]='/', proto:Union[str,Protocol]='https', port:int=None, query:dict=None, username:str=None, password:str=None, anchor:str=None):
		'Create a new `Url` based on the url parts'

		if type(proto) == str:
			try:
				protocol = Protocol[proto.upper()]
			except KeyError:
				protocol = None

		if protocol and port == protocol.value:
			port = None

		url = f'{proto}://'

		if username and password:
			url += f'{username}:{password}@'

		elif username:
			url += f'{username}@'

		url += domain or ''

		if port:
			url += f':{port}'

		url += '/' + path if not path.startswith('/') else path

		if query:
			url += '?' + '&'.join(f'{quote(str(key))}={quote(str(value))}' for key,value in query.items())

		if anchor:
			url += f'#{anchor}'

		return cls(url)


	@cached_property
	def dict(self) -> DotDict:
		'Return the parsed url as a dict'

		return DotDict(
			proto = self.proto,
			domain = self.domain,
			port = self.port,
			path = self.path,
			query = self.query,
			username = self.username,
			password = self.password,
			anchor = self.anchor
		)


	@cached_property
	def address(self) -> str:
		'Resolve the domain to an IP address'

		return self.resolve_hostname()


	@cached_property
	def address_type(self) -> IpAddressType:
		izzylog.warning('Deprecated property. Use `Url.address.type` instead')
		return self.address.type


	@cached_property
	def anchor(self) -> str:
		return self._parsed.fragment


	@cached_property
	def domain(self) -> str:
		return self._parsed.hostname


	@cached_property
	def host(self) -> str:
		'The domain and, if set, port as a string'

		if self.port:
			return f'{self.domain}:{self.port}'

		return self.domain


	@cached_property
	def mimetype(self) -> str:
		'Tries to determine the mimetype based on the file extension'

		return mimetypes.guess_type(self)[0]


	@cached_property
	def password(self) -> str:
		return self._parsed.password


	@cached_property
	def path(self) -> Path:
		return Path(self._parsed.path or '/')


	@cached_property
	def path_full(self) -> str:
		'Return the path with the query pairs and anchor'

		string = self.path

		if self.query_string:
			string += f'?{query_string}'

		if self.anchor:
			string += f'#{self.anchor}'

		return string


	@cached_property
	def port(self) -> int:
		if not self._parsed.port:
			if not self.protocol:
				return

			return self.protocol.value

		else:
			return self._parsed.port


	@cached_property
	def proto(self) -> str:
		return self._parsed.scheme.lower()


	@cached_property
	def protocol(self) -> Protocol:
		try:
			return Protocol[self.proto.upper()]

		except KeyError:
			return None


	@cached_property
	def query_string(self) -> str:
		return self._parsed.query


	@cached_property
	def query(self) -> DotDict:
		return DotDict.new_from_query_string(self._parsed.query)


	@cached_property
	def top(self) -> str:
		'Returns the domain without sub-domains'

		if not self.domain and not self.path:
			raise ValueError('No domain')

		domain = self.domain or self.path

		if not self._tldcache:
			if not self._tldcache_path.exists() or self._tldcache_path.modified() + timedelta(days=7) < datetime.now():
				resp = urlopen('https://publicsuffix.org/list/public_suffix_list.dat')

				with self._tldcache_path.open('w') as fd:
					for line in resp.read().decode('utf-8').splitlines():
						if 'end icann domains' in line.lower():
							break

						if not line or line.startswith('//'):
							continue

						if line.startswith('*'):
							line = line[2:]

						fd.write(line + '\n')

			with self._tldcache_path.open() as fd:
				self._tldcache = set(fd.read().splitlines())

		domain_split = domain.split('.')

		try:
			if '.'.join(domain_split[-2:]) in self._tldcache:
				return '.'.join(domain_split[-3:])

		except IndexError:
			pass

		if '.'.join(domain_split[-1:]) in self._tldcache:
			return '.'.join(domain_split[-2:])

		raise ValueError('Cannot find TLD')


	@cached_property
	def username(self) -> str:
		return self._parsed.username


	@cached_property
	def without_query(self) -> str:
		'Return the url without the query or anchor on the end'

		return Url(self.split('?')[0])


	def copy(self):
		return Url.new(**self.dict)


	def hostname(self, always_return_port=False):
		'Return the domain and port. If always_return_port is false and the port is the default for the protocol, only return the domain.'

		if not always_return_port and (not self.protocol or self.port == self.protocol.value):
			return self.domain

		return self.host


	def is_address_type(self, iptype):
		izzylog.warning('Deprecated property. Use `Url.address.is_type` instead')
		return self.address.is_type(iptype)


	def join(self, new_path):
		'Add to the path portion of the url'

		data = self.dict
		domain = data.pop('domain')

		data['path'] = data['path'].join(new_path)

		return self.new(domain, **data)


	def replace_property(self, key, value):
		return self.replace_properties(**{key: value})


	def replace_properties(self, **kwargs):
		data = self.dict

		for key, value in kwargs.items():
			if key not in data.keys():
				raise KeyError(f'Invalid Url property: {key}')

			data[key] = value

		return self.new(**data)


	def resolve_hostname(self, timeout=5):
		'Resolve the domain to an IP address'

		if not self._address:
			self._address = Address(self.domain, timeout=timeout)

		return self._address
